# sage_setup: distribution = sagemath-graphs
r"""
Top level of the distribution package sagemath-graphs

This distribution makes the following feature available::

    sage: from sage.features.sagemath import *
    sage: sage__graphs().is_present()
    FeatureTestResult('sage.graphs', True)
"""

from .all__sagemath_categories import *

try:  # extra
    from sage.all__sagemath_gap import *
except ImportError:
    pass

try:  # extra
    from sage.all__sagemath_modules import *
except ImportError:
    pass

try:  # extra
    from sage.all__sagemath_polyhedra import *
except ImportError:
    pass

from sage.graphs.all     import *

from sage.topology.all   import *

from sage.combinat.all__sagemath_graphs import *

from sage.databases.all__sagemath_graphs import *

from sage.sandpiles.all import *

from sage.knots.all import *



# sage_setup: distribution = sagemath-graphs
r"""
Incidence structures (i.e. hypergraphs, i.e. set systems)

An incidence structure is specified by a list of points, blocks, or an incidence
matrix ([1]_, [2]_). :class:`IncidenceStructure` instances have the following methods:

{METHODS_OF_IncidenceStructure}

REFERENCES:

.. [1] Block designs and incidence structures from wikipedia,
  :wikipedia:`Block_design`
  :wikipedia:`Incidence_structure`

.. [2] \E. Assmus, J. Key, Designs and their codes, CUP, 1992.

AUTHORS:

- Peter Dobcsanyi and David Joyner (2007-2008)

  This is a significantly modified form of part of the module block_design.py
  (version 0.6) written by Peter Dobcsanyi peter@designtheory.org.

- Vincent Delecroix (2014): major rewrite

Methods
-------
"""
# **************************************************************************
#                              Copyright (C) 2007                          #
#                                                                          #
#                Peter Dobcsanyi       and         David Joyner            #
#           <peter@designtheory.org>          <wdjoyner@gmail.com>         #
#                                                                          #
#                                                                          #
#    Distributed under the terms of the GNU General Public License (GPL)   #
#    as published by the Free Software Foundation; either version 2 of     #
#    the License, or (at your option) any later version.                   #
#                    https://www.gnu.org/licenses/                          #
# **************************************************************************
from __future__ import annotations

from sage.misc.latex import latex
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer
from sage.sets.set import Set

lazy_import('sage.libs.gap.libgap', 'libgap')


class IncidenceStructure:
    r"""
    A base class for incidence structures (i.e. hypergraphs, i.e. set systems)

    An incidence structure (i.e. hypergraph, i.e. set system) can be defined
    from a collection of blocks (i.e. sets, i.e. edges), optionally with an
    explicit ground set (i.e. point set, i.e. vertex set). Alternatively they
    can be defined from a binary incidence matrix.

    INPUT:

    - ``points`` -- (i.e. ground set, i.e. vertex set) the underlying set. If
      ``points`` is an integer `v`, then the set is considered to be `\{0, ...,
      v-1\}`.

      .. NOTE::

          The following syntax, where ``points`` is omitted, automatically
          defines the ground set as the union of the blocks::

              sage: H = IncidenceStructure([['a','b','c'],['c','d','e']])
              sage: sorted(H.ground_set())
              ['a', 'b', 'c', 'd', 'e']

    - ``blocks`` -- (i.e. edges, i.e. sets) the blocks defining the incidence
      structure; can be any iterable

    - ``incidence_matrix`` -- a binary incidence matrix; each column represents
      a set

    - ``name`` -- string (such as "Fano plane")

    - ``check`` -- whether to check the input

    - ``copy`` -- (use with caution) if set to ``False`` then ``blocks`` must be
      a list of lists of integers. The list will not be copied but will be
      modified in place (each block is sorted, and the whole list is
      sorted). Your ``blocks`` object will become the
      :class:`IncidenceStructure` instance's internal data.

    EXAMPLES:

    An incidence structure can be constructed by giving the number of points and
    the list of blocks::

        sage: IncidenceStructure(7, [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
        Incidence structure with 7 points and 7 blocks

    Only providing the set of blocks is sufficient. In this case, the ground set
    is defined as the union of the blocks::

        sage: IncidenceStructure([[1,2,3],[2,3,4]])
        Incidence structure with 4 points and 2 blocks

    Or by its adjacency matrix (a `\{0,1\}`-matrix in which rows are indexed by
    points and columns by blocks)::

        sage: m = matrix([[0,1,0],[0,0,1],[1,0,1],[1,1,1]])                             # needs sage.modules
        sage: IncidenceStructure(m)                                                     # needs sage.modules
        Incidence structure with 4 points and 3 blocks

    The points can be any (hashable) object::

        sage: V = [(0,'a'),(0,'b'),(1,'a'),(1,'b')]
        sage: B = [(V[0],V[1],V[2]), (V[1],V[2]), (V[0],V[2])]
        sage: I = IncidenceStructure(V, B)
        sage: I.ground_set()
        [(0, 'a'), (0, 'b'), (1, 'a'), (1, 'b')]
        sage: I.blocks()
        [[(0, 'a'), (0, 'b'), (1, 'a')], [(0, 'a'), (1, 'a')], [(0, 'b'), (1, 'a')]]

    The order of the points and blocks does not matter as they are sorted on
    input (see :issue:`11333`)::

        sage: A = IncidenceStructure([0,1,2], [[0],[0,2]])
        sage: B = IncidenceStructure([1,0,2], [[0],[2,0]])
        sage: B == A
        True

        sage: C = BlockDesign(2, [[0], [1,0]])
        sage: D = BlockDesign(2, [[0,1], [0]])
        sage: C == D
        True

    If you care for speed, you can set ``copy`` to ``False``, but in that
    case, your input must be a list of lists and the ground set must be `{0,
    ..., v-1}`::

        sage: blocks = [[0,1],[2,0],[1,2]]  # a list of lists of integers
        sage: I = IncidenceStructure(3, blocks, copy=False)
        sage: I._blocks is blocks
        True
    """
    def __init__(self, points=None, blocks=None, incidence_matrix=None,
                 name=None, check=True, copy=True):
        r"""
        TESTS::

            sage: IncidenceStructure(3, [[4]])
            Traceback (most recent call last):
            ...
            ValueError: Block [4] is not contained in the point set

            sage: IncidenceStructure(3, [[0,1],[0,2]], check=True)
            Incidence structure with 3 points and 2 blocks

            sage: IncidenceStructure(2, [[0,1,2,3,4,5]], check=False)
            Incidence structure with 2 points and 1 blocks

        We avoid to convert to integers when the points are not (but compare
        equal to integers because of coercion)::

            sage: # needs sage.rings.finite_rings
            sage: V = GF(5)
            sage: e0,e1,e2,e3,e4 = V
            sage: [e0,e1,e2,e3,e4] == list(range(5))  # coercion makes them equal
            True
            sage: blocks = [[e0,e1,e2],[e0,e1],[e2,e4]]
            sage: I = IncidenceStructure(V, blocks)
            sage: type(I.ground_set()[0])
            <class 'sage.rings.finite_rings.integer_mod.IntegerMod_int'>
            sage: type(I.blocks()[0][0])
            <class 'sage.rings.finite_rings.integer_mod.IntegerMod_int'>

        TESTS::

            sage: IncidenceStructure([])
            Incidence structure with 0 points and 0 blocks
        """
        from sage.structure.element import Matrix

        # Reformatting input
        if isinstance(points, Matrix):
            assert incidence_matrix is None, "'incidence_matrix' cannot be defined when 'points' is a matrix"
            assert blocks is None, "'blocks' cannot be defined when 'points' is a matrix"
            incidence_matrix = points
            points = blocks = None
        elif points is not None and blocks is None:
            blocks = points
            points = set().union(*blocks)
        if points:
            assert incidence_matrix is None, "'incidence_matrix' cannot be defined when 'points' is defined"

        if incidence_matrix:
            from sage.matrix.constructor import matrix
            M = matrix(incidence_matrix)
            v = M.nrows()
            self._points = list(range(v))
            self._point_to_index = None
            self._blocks = sorted(M.nonzero_positions_in_column(i) for i in range(M.ncols()))

        else:
            if isinstance(points, (int, Integer)):
                self._points = list(range(points))
                self._point_to_index = None
            else:
                self._points = list(points)
                if self._points == list(range(len(points))) and all(isinstance(x, (int, Integer)) for x in self._points):
                    self._point_to_index = None
                else:
                    self._point_to_index = {e: i for i, e in enumerate(self._points)}

            if check:
                for block in blocks:
                    if any(x not in self._points for x in block):
                        raise ValueError("Block {} is not contained in the point set".format(block))
                    if len(block) != len(set(block)):
                        raise ValueError("Repeated element in block {}".format(block))

            if self._point_to_index:
                # translate everything to integers between 0 and v-1
                blocks = [sorted(self._point_to_index[e] for e in block) for block in blocks]
            elif copy:
                # create a new list made of sorted blocks
                blocks = [sorted(block) for block in blocks]
            else:
                # sort the data but avoid copying it
                for b in blocks:
                    b.sort()

            blocks.sort()
            self._blocks = blocks

        self._name = str(name) if name is not None else 'IncidenceStructure'
        self._classes = None
        self._canonical_label = None

    def __iter__(self):
        """
        Iterator over the blocks.

        EXAMPLES::

            sage: sts = designs.steiner_triple_system(9)
            sage: list(sts)
            [[0, 1, 5], [0, 2, 4], [0, 3, 6], [0, 7, 8], [1, 2, 3], [1, 4, 7],
            [1, 6, 8], [2, 5, 8], [2, 6, 7], [3, 4, 8], [3, 5, 7], [4, 5, 6]]

            sage: b = IncidenceStructure('ab', ['a','ab'])
            sage: it = iter(b)
            sage: next(it)
            ['a']
            sage: next(it)
            ['a', 'b']
        """
        if self._point_to_index is None:
            for b in self._blocks:
                yield b[:]
        else:
            for b in self._blocks:
                yield [self._points[i] for i in b]

    def __repr__(self):
        """
        A print method.

        EXAMPLES::

            sage: BD = IncidenceStructure(7,[[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
            sage: BD
            Incidence structure with 7 points and 7 blocks
        """
        return 'Incidence structure with {} points and {} blocks'.format(
            self.num_points(), self.num_blocks())

    __str__ = __repr__

    def __eq__(self, other):
        """
        Test whether the two incidence structures are equal.

        TESTS::

            sage: blocks = [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]]
            sage: BD1 = IncidenceStructure(7, blocks)
            sage: M = BD1.incidence_matrix()                                            # needs sage.modules
            sage: BD2 = IncidenceStructure(incidence_matrix=M)                          # needs sage.modules
            sage: BD1 == BD2                                                            # needs sage.modules
            True

            sage: e1 = frozenset([0,1])
            sage: e2 = frozenset([2])
            sage: sorted([e1,e2]) == [e1,e2]
            True
            sage: sorted([e2,e1]) == [e2,e1]
            True
            sage: I1 = IncidenceStructure([e1,e2], [[e1],[e1,e2]])
            sage: I2 = IncidenceStructure([e1,e2], [[e2,e1],[e1]])
            sage: I3 = IncidenceStructure([e2,e1], [[e1,e2],[e1]])
            sage: I1 == I2 and I2 == I1 and I1 == I3 and I3 == I1 and I2 == I3 and I3 == I2
            True
        """
        # We are extra careful in this method since we cannot assume that a
        # total order is defined on the point set.
        if not isinstance(other, IncidenceStructure):
            return False

        if self._points == other._points:
            return self._blocks == other._blocks

        if (self.num_points() != other.num_points() or
                self.num_blocks() != other.num_blocks()):
            return False

        p_to_i = self._point_to_index if self._point_to_index else list(range(self.num_points()))

        if any(p not in p_to_i for p in other.ground_set()):
            return False

        other_blocks = sorted(sorted(p_to_i[p] for p in b) for b in other.blocks())
        return self._blocks == other_blocks

    def __ne__(self, other):
        r"""
        Difference test.

        EXAMPLES::

            sage: BD1 = IncidenceStructure(7, [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
            sage: M = BD1.incidence_matrix()                                            # needs sage.modules
            sage: BD2 = IncidenceStructure(incidence_matrix=M)                          # needs sage.modules
            sage: BD1 != BD2                                                            # needs sage.modules
            False
        """
        return not self == other

    def __contains__(self, block):
        r"""
        Test if a block belongs to the incidence structure.

        INPUT:

        - ``block`` -- a block

        EXAMPLES::

            sage: [1,2,3,4] in IncidenceStructure([[1,2,3,4]])
            True
            sage: [1,2,4,3] in IncidenceStructure([[1,2,3,4]])
            True
            sage: [1,2,"3",4] in IncidenceStructure([[1,2,3,4]])
            False
            sage: [1,2,"3",4] in IncidenceStructure([[1,2,"3",4]])
            True

        More complicated examples::

            sage: str="I had a dream of a time when a 3-lines patch does not kill one hour"
            sage: sets = Subsets(str.split(), 4)
            sage: IS = IncidenceStructure(sets)  # a complete 4-uniform hypergraph
            sage: ["I", "dream", "of", "one"] in IS
            True
            sage: ["does", "patch", "kill", "dream"] in IS
            True
            sage: ["Am", "I", "finally", "done ?"] in IS
            False
            sage: IS = designs.ProjectiveGeometryDesign(3, 1, GF(2),                    # needs sage.combinat sage.modules
            ....:                                       point_coordinates=False)
            sage: [3,8,7] in IS                                                         # needs sage.combinat sage.modules
            True
            sage: [3,8,9] in IS                                                         # needs sage.combinat sage.modules
            False
        """
        try:
            iter(block)
        except TypeError:
            return False

        # Relabel to 0,...,n-1 if necessary
        if self._point_to_index is not None:
            try:
                block = [self._point_to_index[x] for x in block]
            except KeyError:
                return False

        return sorted(block) in self._blocks

    def canonical_label(self):
        r"""
        Return a canonical label for the incidence structure.

        A canonical label is relabeling of the points into integers
        `\{0,...,n-1\}` such that isomorphic incidence structures are
        relabelled to equal objects.

        EXAMPLES::

            sage: # needs sage.schemes
            sage: fano1 = designs.balanced_incomplete_block_design(7,3)
            sage: fano2 = designs.projective_plane(2)
            sage: fano1 == fano2
            False
            sage: fano1.relabel(fano1.canonical_label())
            sage: fano2.relabel(fano2.canonical_label())
            sage: fano1 == fano2
            True
        """
        if self._canonical_label is None:
            from sage.graphs.graph import Graph
            g = Graph()
            n = self.num_points()
            g.add_edges((i+n, x) for i, b in enumerate(self._blocks) for x in b)
            canonical_label = g.canonical_label([list(range(n)), list(range(n, n+self.num_blocks()))], certificate=True)[1]
            canonical_label = [canonical_label[x] for x in range(n)]
            self._canonical_label = canonical_label

        return dict(zip(self._points, self._canonical_label))

    def is_isomorphic(self, other, certificate=False):
        r"""
        Return whether the two incidence structures are isomorphic.

        INPUT:

        - ``other`` -- an incidence structure

        - ``certificate`` -- boolean (default: ``False``); whether to return an
          isomorphism from ``self`` to ``other`` instead of a boolean answer

        EXAMPLES::

            sage: # needs sage.schemes
            sage: fano1 = designs.balanced_incomplete_block_design(7,3)
            sage: fano2 = designs.projective_plane(2)
            sage: fano1.is_isomorphic(fano2)
            True
            sage: fano1.is_isomorphic(fano2,certificate=True)
            {0: 0, 1: 1, 2: 2, 3: 6, 4: 4, 5: 3, 6: 5}

        TESTS::

            sage: # needs sage.symbolic
            sage: IS  = IncidenceStructure([["A",5,pi],["A",5,"Wouhou"],
            ....:                           ["A","Wouhou",(9,9)],[pi,12]])
            sage: IS2 = IS.copy()
            sage: IS2.relabel(IS2.canonical_label())
            sage: IS.is_isomorphic(IS2)
            True
            sage: canon = IS.is_isomorphic(IS2, certificate=True)
            sage: IS.relabel(canon)
            sage: IS==IS2
            True

            sage: IS2 = IncidenceStructure([[1,2]])
            sage: IS2.is_isomorphic(IS)                                                 # needs sage.symbolic
            False
            sage: IS2.is_isomorphic(IS, certificate=True)                               # needs sage.symbolic
            {}

        Checking whether two :class:`IncidenceStructure` are isomorphic
        incidentally computes their canonical label (if necessary). Thus,
        subsequent calls to :meth:`is_isomorphic` will be faster::

            sage: # needs sage.schemes
            sage: IS1 = designs.projective_plane(3)
            sage: IS2 = IS1.relabel(Permutations(IS1.ground_set()).random_element(),
            ....:                   inplace=False)
            sage: IS2 = IncidenceStructure(IS2.blocks())
            sage: IS1._canonical_label is None and IS2._canonical_label is None
            True
            sage: IS1.is_isomorphic(IS2)
            True
            sage: IS1._canonical_label is None or IS2._canonical_label is None
            False
        """
        if (self.num_points() != other.num_points() or
            self.num_blocks() != other.num_blocks() or
                sorted(self.block_sizes()) != sorted(other.block_sizes())):
            return {} if certificate else False

        A_canon = self.canonical_label()
        B_canon = other.canonical_label()

        A = self.relabel(A_canon, inplace=False)
        B = other.relabel(B_canon, inplace=False)

        if A == B:
            if certificate:
                B_canon_rev = {y: x for x, y in B_canon.items()}
                return {x: B_canon_rev[xint] for x, xint in A_canon.items()}
            else:
                return True
        else:
            return {} if certificate else False

    def isomorphic_substructures_iterator(self, H2, induced=False):
        r"""
        Iterate over all copies of ``H2`` contained in ``self``.

        A hypergraph `H_1` contains an isomorphic copy of a hypergraph `H_2` if
        there exists an injection `f:V(H_2)\mapsto V(H_1)` such that for any set
        `S_2\in E(H_2)` the set `S_1=f(S2)` belongs to `E(H_1)`.

        It is an *induced* copy if no other set of `E(H_1)` is contained in
        `f(V(H_2))`, i.e. `|E(H_2)|=\{S:S\in E(H_1)\text{ and }f(V(H_2))\}`.

        This function lists all such injections. In particular, the number of
        copies of `H` in itself is equal to *the size of its automorphism
        group*.

        See :mod:`~sage.combinat.designs.subhypergraph_search` for more information.

        INPUT:

        - ``H2`` -- an :class:`IncidenceStructure` object

        - ``induced`` -- boolean (default: ``False``); whether to require the copies to be
          induced

        EXAMPLES:

        How many distinct `C_5` in Petersen's graph ? ::

            sage: P = graphs.PetersenGraph()
            sage: C = graphs.CycleGraph(5)
            sage: IP = IncidenceStructure(P.edges(sort=True, labels=False))
            sage: IC = IncidenceStructure(C.edges(sort=True, labels=False))
            sage: sum(1 for _ in IP.isomorphic_substructures_iterator(IC))
            120

        As the automorphism group of `C_5` has size 10, the number of distinct
        unlabelled copies is 12. Let us check that all functions returned
        correspond to an actual `C_5` subgraph::

            sage: for f in IP.isomorphic_substructures_iterator(IC):
            ....:     assert all(P.has_edge(f[x],f[y]) for x,y in C.edges(sort=True, labels=False))

        The number of induced copies, in this case, is the same::

            sage: sum(1 for _ in IP.isomorphic_substructures_iterator(IC,induced=True))
            120

        They begin to differ if we make one vertex universal::

            sage: P.add_edges([(0,x) for x in P], loops=False)
            sage: IP = IncidenceStructure(P.edges(sort=True, labels=False))
            sage: IC = IncidenceStructure(C.edges(sort=True, labels=False))
            sage: sum(1 for _ in IP.isomorphic_substructures_iterator(IC))
            420
            sage: sum(1 for _ in IP.isomorphic_substructures_iterator(IC,induced=True))
            60

        The number of copies of `H` in itself is the size of its automorphism
        group::

            sage: H = designs.projective_plane(3)                                       # needs sage.schemes
            sage: sum(1 for _ in H.isomorphic_substructures_iterator(H))                # needs sage.schemes
            5616
            sage: H.automorphism_group().cardinality()                                  # needs sage.groups sage.schemes
            5616
        """
        from sage.combinat.designs.subhypergraph_search import SubHypergraphSearch
        return SubHypergraphSearch(self, H2, induced=induced)

    def copy(self):
        r"""
        Return a copy of the incidence structure.

        EXAMPLES::

            sage: IS = IncidenceStructure([[1,2,3,"e"]], name='Test')
            sage: IS
            Incidence structure with 4 points and 1 blocks
            sage: copy(IS)
            Incidence structure with 4 points and 1 blocks
            sage: [1, 2, 3, 'e'] in copy(IS)
            True
            sage: copy(IS)._name
            'Test'
        """
        IS = IncidenceStructure(self._blocks,
                                name=self._name,
                                check=False)
        IS.relabel(dict(zip(range(self.num_points()), self._points)))
        IS._canonical_label = None if self._canonical_label is None else self._canonical_label[:]

        return IS

    __copy__ = copy

    def induced_substructure(self, points):
        r"""
        Return the substructure induced by a set of points.

        The substructure induced in `\mathcal H` by a set `X\subseteq V(\mathcal
        H)` of points is the incidence structure `\mathcal H_X` defined on `X`
        whose sets are all `S\in \mathcal H` such that `S\subseteq X`.

        INPUT:

        - ``points`` -- set of points

        .. NOTE::

            This method goes over all sets of ``self`` before building a new
            :class:`IncidenceStructure` (which involves some relabelling and
            sorting). It probably should not be called in a performance-critical
            code.

        EXAMPLES:

        A Fano plane with one point removed::

            sage: F = designs.steiner_triple_system(7)
            sage: F.induced_substructure([0..5])
            Incidence structure with 6 points and 4 blocks

        TESTS::

            sage: F.induced_substructure([0..50])
            Traceback (most recent call last):
            ...
            ValueError: 7 is not a point of the incidence structure
            sage: F.relabel(dict(enumerate("abcdefg")))
            sage: F.induced_substructure("abc")
            Incidence structure with 3 points and ...
            sage: F.induced_substructure("Y")
            Traceback (most recent call last):
            ...
            ValueError: 'Y' is not a point of the incidence structure
        """
        # Checking the input
        if self._point_to_index is None:
            n = self.num_points()
            for x in points:
                x = int(x)
                if x < 0 or x >= n:
                    raise ValueError("{} is not a point of the incidence structure".format(x))
            int_points = points
        else:
            try:
                int_points = [self._point_to_index[x] for x in points]
            except KeyError as bad_pt:
                raise ValueError("{} is not a point of the incidence structure".format(bad_pt))

        int_points = set(int_points)
        return IncidenceStructure(points,
                                  [[self._points[x] for x in S]
                                   for S in self._blocks
                                   if int_points.issuperset(S)])

    def trace(self, points, min_size=1, multiset=True):
        r"""
        Return the trace of a set of points.

        Given an hypergraph `\mathcal H`, the *trace* of a set `X` of points in
        `\mathcal H` is the hypergraph whose blocks are all non-empty `S \cap X`
        where `S \in \mathcal H`.

        INPUT:

        - ``points`` -- set of points

        - ``min_size`` -- integer (default: 1); minimum size of the sets to
          keep. By default all empty sets are discarded, i.e. ``min_size=1``

        - ``multiset`` -- boolean (default: ``True``); whether to keep multiple
          copies of the same set

        .. NOTE::

            This method goes over all sets of ``self`` before building a new
            :class:`IncidenceStructure` (which involves some relabelling and
            sorting). It probably should not be called in a performance-critical
            code.

        EXAMPLES:

        A Baer subplane of order 2 (i.e. a Fano plane) in a projective plane of order 4::

            sage: # needs sage.schemes
            sage: P4 = designs.projective_plane(4)
            sage: F = designs.projective_plane(2)
            sage: for x in Subsets(P4.ground_set(),7):
            ....:     if P4.trace(x,min_size=2).is_isomorphic(F):
            ....:         break
            sage: subplane = P4.trace(x,min_size=2); subplane
            Incidence structure with 7 points and 7 blocks
            sage: subplane.is_isomorphic(F)
            True

        TESTS::

            sage: # needs sage.schemes
            sage: F.trace([0..50])
            Traceback (most recent call last):
            ...
            ValueError: 7 is not a point of the incidence structure
            sage: F.relabel(dict(enumerate("abcdefg")))
            sage: F.trace("abc")
            Incidence structure with 3 points and ...
            sage: F.trace("Y")
            Traceback (most recent call last):
            ...
            ValueError: 'Y' is not a point of the incidence structure
        """
        # Checking the input
        if self._point_to_index is None:
            n = self.num_points()
            int_points = frozenset(int(x) for x in points)
            for x in int_points:
                if x < 0 or x >= n:
                    raise ValueError("{} is not a point of the incidence structure".format(x))
        else:
            try:
                int_points = frozenset(self._point_to_index[x] for x in points)
            except KeyError as bad_pt:
                raise ValueError("{} is not a point of the incidence structure".format(bad_pt))

        blocks = [int_points.intersection(S) for S in self._blocks]
        if min_size:
            blocks = [S for S in blocks if len(S) >= min_size]
        if not multiset:
            blocks = set(blocks)
        IS = IncidenceStructure(blocks)
        IS.relabel({i: self._points[i] for i in int_points})
        return IS

    def ground_set(self):
        r"""
        Return the ground set (i.e the list of points).

        EXAMPLES::

            sage: IncidenceStructure(3, [[0,1],[0,2]]).ground_set()
            [0, 1, 2]
        """
        return self._points[:]

    def num_points(self):
        r"""
        Return the size of the ground set.

        EXAMPLES::

            sage: designs.DesarguesianProjectivePlaneDesign(2).num_points()
            7
            sage: B = IncidenceStructure(4, [[0,1],[0,2],[0,3],[1,2], [1,2,3]])
            sage: B.num_points()
            4
        """
        return len(self._points)

    def num_blocks(self):
        r"""
        Return the number of blocks.

        EXAMPLES::

            sage: designs.DesarguesianProjectivePlaneDesign(2).num_blocks()
            7
            sage: B = IncidenceStructure(4, [[0,1],[0,2],[0,3],[1,2], [1,2,3]])
            sage: B.num_blocks()
            5
        """
        return len(self._blocks)

    def blocks(self):
        """
        Return the list of blocks.

        EXAMPLES::

            sage: BD = IncidenceStructure(7,[[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
            sage: BD.blocks()
            [[0, 1, 2], [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6], [2, 3, 6], [2, 4, 5]]
        """
        if self._point_to_index is None:
            return [b[:] for b in self._blocks]
        else:
            return [[self._points[i] for i in b] for b in self._blocks]

    def block_sizes(self):
        r"""
        Return the set of block sizes.

        EXAMPLES::

            sage: BD = IncidenceStructure(8, [[0,1,3],[1,4,5,6],[1,2],[5,6,7]])
            sage: BD.block_sizes()
            [3, 2, 4, 3]
            sage: BD = IncidenceStructure(7,[[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
            sage: BD.block_sizes()
            [3, 3, 3, 3, 3, 3, 3]
        """
        return [len(b) for b in self._blocks]

    def degree(self, p=None, subset=False):
        r"""
        Return the degree of a point ``p`` (or a set of points).

        The degree of a point (or set of points) is the number of blocks that
        contain it.

        INPUT:

        - ``p`` -- a point (or a set of points) of the incidence structure

        - ``subset`` -- boolean (default: ``False``); whether to interpret the
          argument as a set of point or as a point (default)

        EXAMPLES::

            sage: designs.steiner_triple_system(9).degree(3)
            4
            sage: designs.steiner_triple_system(9).degree({1,2},subset=True)
            1

        TESTS::

            sage: designs.steiner_triple_system(9).degree(subset=True)
            Traceback (most recent call last):
            ...
            ValueError: subset must be False when p is None
        """
        if p is None:
            if subset is True:
                raise ValueError("subset must be False when p is None")

        # degree of a point
        if not subset:
            if self._point_to_index:
                p = self._point_to_index.get(p, -1)
            else:
                p = p if (p >= 0 and p < len(self._points)) else -1
            return sum((p in b) for b in self._blocks) if p != -1 else 0

        # degree of a set
        else:
            if self._point_to_index:
                p = set(self._point_to_index.get(x, -1) for x in p)
            else:
                p = set(p) if all(x >= 0 and x < len(self._points) for x in p) else set([-1])

            return sum(p.issubset(b) for b in self._blocks) if -1 not in p else 0

    def degrees(self, size=None):
        r"""
        Return the degree of all sets of given size, or the degree of all points.

        The degree of a point (or set of point) is the number of blocks that
        contain it.

        INPUT:

        - ``size`` -- integer; return the degree of all subsets of points of
          cardinality ``size``. When ``size=None``, the function outputs the
          degree of all points.

          .. NOTE::

              When ``size=None`` the output is indexed by the points. When
              ``size=1`` it is indexed by tuples of size 1. This is the same
              information, stored slightly differently.

        OUTPUT: a dictionary whose values are degrees and keys are either:

        - the points of the incidence structure if ``size=None`` (default)

        - the subsets of size ``size`` of the points stored as tuples

        EXAMPLES::

            sage: IncidenceStructure([[1,2,3],[1,4]]).degrees(2)
            {(1, 2): 1, (1, 3): 1, (1, 4): 1, (2, 3): 1, (2, 4): 0, (3, 4): 0}

        In a Steiner triple system, all pairs have degree 1::

            sage: S13 = designs.steiner_triple_system(13)
            sage: all(v == 1 for v in S13.degrees(2).values())
            True
        """
        if size is None:
            d = [0]*self.num_points()
            for b in self._blocks:
                for x in b:
                    d[x] += 1
            return {p: d[i] for i, p in enumerate(self._points)}
        else:
            from itertools import combinations
            d = {t: 0 for t in combinations(range(self.num_points()), size)}
            for b in self._blocks:
                for s in combinations(b, size):
                    d[s] += 1
            if self._point_to_index:
                return {tuple([self._points[x] for x in s]): v for s, v in d.items()}
            else:
                return d

    def rank(self):
        r"""
        Return the rank of the hypergraph (the maximum size of a block).

        EXAMPLES::

            sage: h = Hypergraph(8, [[0,1,3],[1,4,5,6],[1,2]])
            sage: h.rank()
            4
        """
        return max(len(b) for b in self._blocks)

    def is_regular(self, r=None) -> bool | int:
        r"""
        Test whether the incidence structure is `r`-regular.

        An incidence structure is said to be `r`-regular if all its points are
        incident with exactly `r` blocks.

        INPUT:

        - ``r`` -- integer

        OUTPUT:

        If ``r`` is defined, a boolean is returned. If ``r`` is set to ``None``
        (default), the method returns either ``False`` or the integer ``r`` such
        that the incidence structure is `r`-regular.

        .. WARNING::

            In case of `0`-regular incidence structure, beware that ``if not
            H.is_regular()`` is a satisfied condition.

        EXAMPLES::

            sage: designs.balanced_incomplete_block_design(7,3).is_regular()            # needs sage.schemes
            3
            sage: designs.balanced_incomplete_block_design(7,3).is_regular(r=3)         # needs sage.schemes
            True
            sage: designs.balanced_incomplete_block_design(7,3).is_regular(r=4)         # needs sage.schemes
            False

        TESTS::

            sage: IncidenceStructure([]).is_regular()
            Traceback (most recent call last):
            ...
            ValueError: This incidence structure has no points.
        """
        if self.num_points() == 0:
            raise ValueError("This incidence structure has no points.")
        count = [0] * self.num_points()
        for b in self._blocks:
            for x in b:
                count[x] += 1
        scount = set(count)
        if len(scount) != 1:
            return False
        if r is None:
            return scount.pop()
        return scount.pop() == r

    def is_uniform(self, k=None) -> bool | int:
        r"""
        Test whether the incidence structure is `k`-uniform

        An incidence structure is said to be `k`-uniform if all its blocks have
        size `k`.

        INPUT:

        - ``k`` -- integer

        OUTPUT:

        If ``k`` is defined, a boolean is returned. If ``k`` is set to ``None``
        (default), the method returns either ``False`` or the integer ``k`` such
        that the incidence structure is `k`-uniform.

        .. WARNING::

            In case of `0`-uniform incidence structure, beware that ``if not
            H.is_uniform()`` is a satisfied condition.

        EXAMPLES::

            sage: designs.balanced_incomplete_block_design(7,3).is_uniform()            # needs sage.schemes
            3
            sage: designs.balanced_incomplete_block_design(7,3).is_uniform(k=3)         # needs sage.schemes
            True
            sage: designs.balanced_incomplete_block_design(7,3).is_uniform(k=4)         # needs sage.schemes
            False

        TESTS::

            sage: IncidenceStructure([]).is_uniform()
            Traceback (most recent call last):
            ...
            ValueError: This incidence structure has no blocks.
        """
        if self.num_blocks() == 0:
            raise ValueError("This incidence structure has no blocks.")
        sizes = set(self.block_sizes())
        if len(sizes) != 1:
            return False
        if k is None:
            return sizes.pop()
        return sizes.pop() == k

    def is_connected(self) -> bool:
        r"""
        Test whether the design is connected.

        EXAMPLES::

            sage: IncidenceStructure(3, [[0,1],[0,2]]).is_connected()
            True
            sage: IncidenceStructure(4, [[0,1],[2,3]]).is_connected()
            False
        """
        from sage.sets.disjoint_set import DisjointSet
        D = DisjointSet(self.num_points())
        for B in self._blocks:
            x = B[0]
            for i in range(1, len(B)):
                D.union(x, B[i])
        return D.number_of_subsets() == 1

    def is_simple(self) -> bool:
        r"""
        Test whether this design is simple (i.e. no repeated block).

        EXAMPLES::

            sage: IncidenceStructure(3, [[0,1],[1,2],[0,2]]).is_simple()
            True
            sage: IncidenceStructure(3, [[0],[0]]).is_simple()
            False

            sage: V = [(0,'a'),(0,'b'),(1,'a'),(1,'b')]
            sage: B = [[V[0],V[1]], [V[1],V[2]]]
            sage: I = IncidenceStructure(V, B)
            sage: I.is_simple()
            True
            sage: I2 = IncidenceStructure(V, B*2)
            sage: I2.is_simple()
            False
        """
        B = self._blocks
        return all(B[i] != B[i + 1] for i in range(len(B) - 1))

    def _gap_(self):
        """
        Return the GAP string describing the design.

        EXAMPLES::

            sage: BD = IncidenceStructure(7,[[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
            sage: BD._gap_()
            'BlockDesign(7,[[1, 2, 3], [1, 4, 5], [1, 6, 7], [2, 4, 6], [2, 5, 7], [3, 4, 7], [3, 5, 6]])'
        """
        v = self.num_points()
        gB = [[x + 1 for x in b] for b in self._blocks]
        return "BlockDesign({},{})".format(v, gB)

    def _libgap_(self):
        """
        Return the design as a GAP record.

        EXAMPLES::

            sage: D = IncidenceStructure(4, [[0,2],[1,2,3],[2,3]])
            sage: D._libgap_()                # optional - gap_package_design
            rec( blocks := [ [ 1, 3 ], [ 2, 3, 4 ], [ 3, 4 ] ],
            isBlockDesign := true, v := 4 )
        """
        libgap.load_package("design")
        v = self.num_points()
        gB = [[x + 1 for x in b] for b in self._blocks]
        return libgap.BlockDesign(v, gB)

    def intersection_graph(self, sizes=None):
        r"""
        Return the intersection graph of the incidence structure.

        The vertices of this graph are the :meth:`blocks` of the incidence
        structure. Two of them are adjacent if the size of their intersection
        belongs to the set ``sizes``.

        INPUT:

        - ``sizes`` -- list/set of integers; for convenience, setting
          ``sizes`` to ``5`` has the same effect as ``sizes=[5]``. When set to
          ``None`` (default), behaves as ``sizes=PositiveIntegers()``.

        EXAMPLES:

        The intersection graph of a
        :func:`~sage.combinat.designs.bibd.balanced_incomplete_block_design` is
        a :meth:`strongly regular graph <Graph.is_strongly_regular>` (when it is
        not trivial)::

            sage: BIBD =  designs.balanced_incomplete_block_design(19,3)
            sage: G = BIBD.intersection_graph(1)
            sage: G.is_strongly_regular(parameters=True)
            (57, 24, 11, 9)
        """
        from sage.sets.positive_integers import PositiveIntegers
        from sage.graphs.graph import Graph
        from sage.sets.set import Set
        if sizes is None:
            sizes = PositiveIntegers()
        elif sizes in PositiveIntegers():
            sizes = (sizes,)
        V = [Set(v) for v in self]
        return Graph([V, lambda x, y: len(x & y) in sizes], loops=False)

    def incidence_matrix(self):
        r"""
        Return the incidence matrix `A` of the design. A is a `(v \times b)`
        matrix defined by: ``A[i,j] = 1`` if ``i`` is in block ``B_j`` and 0
        otherwise.

        EXAMPLES::

            sage: BD = IncidenceStructure(7, [[0,1,2],[0,3,4],[0,5,6],[1,3,5],
            ....:                             [1,4,6],[2,3,6],[2,4,5]])
            sage: BD.block_sizes()
            [3, 3, 3, 3, 3, 3, 3]
            sage: BD.incidence_matrix()                                                 # needs sage.modules
            [1 1 1 0 0 0 0]
            [1 0 0 1 1 0 0]
            [1 0 0 0 0 1 1]
            [0 1 0 1 0 1 0]
            [0 1 0 0 1 0 1]
            [0 0 1 1 0 0 1]
            [0 0 1 0 1 1 0]

            sage: I = IncidenceStructure('abc', ('ab','abc','ac','c'))
            sage: I.incidence_matrix()                                                  # needs sage.modules
            [1 1 1 0]
            [1 1 0 0]
            [0 1 1 1]
        """
        from sage.matrix.constructor import matrix
        from sage.rings.integer_ring import ZZ
        A = matrix(ZZ, self.num_points(), self.num_blocks(), sparse=True)
        for j, b in enumerate(self._blocks):
            for i in b:
                A[i, j] = 1
        return A

    def incidence_graph(self, labels=False):
        r"""
        Return the incidence graph of the incidence structure.

        A point and a block are adjacent in this graph whenever they are
        incident.

        INPUT:

        - ``labels`` -- boolean; whether to return a graph whose vertices are
          integers, or labelled elements

          - ``labels is False`` -- default; in this case the first vertices
            of the graphs are the elements of :meth:`ground_set`, and appear
            in the same order. Similarly, the following vertices represent the
            elements of :meth:`blocks`, and appear in the same order.

          - ``labels is True``, the points keep their original labels, and the
            blocks are :func:`Set <Set>` objects.

            Note that the labelled incidence graph can be incorrect when
            blocks are repeated, and on some (rare) occasions when the
            elements of :meth:`ground_set` mix :func:`Set` and non-:func:`Set
            <Set>` objects.

        EXAMPLES::

            sage: BD = IncidenceStructure(7, [[0,1,2],[0,3,4],[0,5,6],[1,3,5],
            ....:                             [1,4,6],[2,3,6],[2,4,5]])
            sage: BD.incidence_graph()                                                  # needs sage.modules
            Bipartite graph on 14 vertices
            sage: A = BD.incidence_matrix()                                             # needs sage.modules
            sage: Graph(block_matrix([[A*0, A],                                         # needs sage.modules
            ....:                     [A.transpose(),A*0]])) == BD.incidence_graph()
            True

        TESTS:

        With ``labels = True``::

            sage: BD.incidence_graph(labels=True).has_edge(0,Set([0,1,2]))
            True
        """
        if labels:
            from sage.graphs.graph import Graph
            from sage.sets.set import Set
            G = Graph()
            G.add_vertices(self.ground_set())
            for b in self.blocks():
                b = Set(b)
                G.add_vertex(b)
                G.add_edges((b, x) for x in b)
            return G

        else:
            from sage.graphs.bipartite_graph import BipartiteGraph
            A = self.incidence_matrix()
            return BipartiteGraph(A)

    def is_berge_cyclic(self):
        r"""
        Check whether ``self`` is a Berge-Cyclic uniform hypergraph.

        A `k`-uniform Berge cycle (named after Claude Berge) of length `\ell`
        is a cyclic list of distinct `k`-sets `F_1,\ldots,F_\ell`, `\ell>1`,
        and distinct vertices `C = \{v_1,\ldots,v_\ell\}` such that for each
        `1\le i\le \ell`, `F_i` contains `v_i` and `v_{i+1}` (where `v_{l+1} =
        v_1`).

        A uniform hypergraph is Berge-cyclic if its incidence graph is cyclic.
        It is called "Berge-acyclic" otherwise.

        For more information, see [Fag1983]_ and :wikipedia:`Hypergraph`.

        EXAMPLES::

            sage: Hypergraph(5, [[1, 2, 3], [2, 3, 4]]).is_berge_cyclic()               # needs sage.modules
            True
            sage: Hypergraph(6, [[1, 2, 3], [3, 4, 5]]).is_berge_cyclic()               # needs sage.modules
            False

        TESTS::

            sage: Hypergraph(5, [[1, 2, 3], [2, 3]]).is_berge_cyclic()
            Traceback (most recent call last):
            ...
            TypeError: Berge cycles are defined for uniform hypergraphs only
        """
        if not self.is_uniform():
            raise TypeError("Berge cycles are defined for uniform hypergraphs only")

        return not self.incidence_graph().is_forest()

    def complement(self, uniform=False):
        r"""
        Return the complement of the incidence structure.

        Two different definitions of "complement" are made available, according
        to the value of ``uniform``.

        INPUT:

        - ``uniform`` -- boolean

          - if set to ``False`` (default), returns the incidence structure whose
            blocks are the complements of all blocks of the incidence structure.

          - If set to ``True`` and the incidence structure is `k`-uniform,
            returns the incidence structure whose blocks are all `k`-sets of the
            ground set that do not appear in ``self``.

        EXAMPLES:

        The complement of a
        :class:`~sage.combinat.designs.bibd.BalancedIncompleteBlockDesign` is
        also a `2`-design::

            sage: bibd = designs.balanced_incomplete_block_design(13,4)                 # needs sage.schemes
            sage: bibd.is_t_design(return_parameters=True)                              # needs sage.schemes
            (True, (2, 13, 4, 1))
            sage: bibd.complement().is_t_design(return_parameters=True)                 # needs sage.schemes
            (True, (2, 13, 9, 6))

        The "uniform" complement of a graph is a graph::

            sage: g = graphs.PetersenGraph()
            sage: G = IncidenceStructure(g.edges(sort=True, labels=False))
            sage: H = G.complement(uniform=True)
            sage: h = Graph(H.blocks())
            sage: g == h
            False
            sage: g == h.complement()
            True

        TESTS::

            sage: bibd.relabel({i:str(i) for i in bibd.ground_set()})                   # needs sage.schemes
            sage: bibd.complement().ground_set()                                        # needs sage.schemes
            ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12']

            sage: I = IncidenceStructure('abc', ['ab','ac','bc'])
            sage: I.is_t_design(return_parameters=True)
            (True, (2, 3, 2, 1))
        """
        if uniform:
            k = self.is_uniform()
            if k is False:
                raise ValueError("The incidence structure is not uniform.")

            blocks = []
            num_blocks = self.num_blocks()
            i = 0
            from itertools import combinations
            for B in combinations(range(self.num_points()), k):
                B = list(B)
                while i < num_blocks and self._blocks[i] < B:
                    i += 1
                if i < num_blocks and self._blocks[i] == B:
                    i += 1
                    continue
                blocks.append(B)
            I = IncidenceStructure(blocks, copy=False)
        else:
            X = set(range(self.num_points()))
            I = IncidenceStructure([X.difference(B) for B in self._blocks])

        I.relabel({i: self._points[i] for i in range(self.num_points())})
        return I

    def relabel(self, perm=None, inplace=True):
        r"""
        Relabel the ground set.

        INPUT:

        - ``perm`` -- can be one of

            - a dictionary -- then each point ``p`` (which should be a key of
              ``d``) is relabeled to ``d[p]``

            - a list or a tuple of length ``n`` -- the first point returned by
              :meth:`ground_set` is relabeled to ``l[0]``, the second to
              ``l[1]``, ...

            - ``None`` -- the incidence structure is relabeled to be on
              `\{0,1,...,n-1\}` in the ordering given by :meth:`ground_set`

        - ``inplace`` -- boolean (default: ``False``); if ``True`` then return
          a relabeled graph and does not touch ``self``

        EXAMPLES::

            sage: # needs sage.schemes
            sage: TD = designs.transversal_design(5,5)
            sage: TD.relabel({i: chr(97+i) for i in range(25)})
            sage: TD.ground_set()
            ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
             'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y']
            sage: TD.blocks()[:3]
            [['a', 'f', 'k', 'p', 'u'], ['a', 'g', 'm', 's', 'y'], ['a', 'h', 'o', 'q', 'x']]

        Relabel to integer points::

            sage: TD.relabel()                                                          # needs sage.schemes
            sage: TD.blocks()[:3]                                                       # needs sage.schemes
            [[0, 5, 10, 15, 20], [0, 6, 12, 18, 24], [0, 7, 14, 16, 23]]

        TESTS:

        Check that the relabel is consistent on a fixed incidence structure::

            sage: I = IncidenceStructure([0,1,2,3,4],
            ....:               [[0,1,3],[0,2,4],[2,3,4],[0,1]])
            sage: I.relabel()
            sage: from itertools import permutations
            sage: for p in permutations([0,1,2,3,4]):
            ....:     J = I.relabel(p,inplace=False)
            ....:     if I == J: print(p)
            (0, 1, 2, 3, 4)
            (0, 1, 4, 3, 2)

        And one can also verify that we have exactly two automorphisms::

            sage: I.automorphism_group()                                                # needs sage.groups
            Permutation Group with generators [(2,4)]
        """
        if not inplace:
            from copy import copy
            G = copy(self)
            G.relabel(perm=perm, inplace=True)
            return G

        if perm is None:
            self._points = list(range(self.num_points()))
            self._point_to_index = None
            return

        if isinstance(perm, (list, tuple)):
            perm = dict(zip(self._points, perm))

        if not isinstance(perm, dict):
            raise ValueError("perm argument must be None, a list or a dictionary")

        if len(set(perm.values())) != len(perm):
            raise ValueError("two points are getting relabelled with the same name")

        self._points = [perm[x] for x in self._points]
        if self._points == list(range(self.num_points())):
            self._point_to_index = None
        else:
            self._point_to_index = {v: i for i, v in enumerate(self._points)}

    # __hash__ = None
    # This object is mutable because of .relabel()

    #####################
    # real computations #
    #####################

    def packing(self, solver=None, verbose=0, *, integrality_tolerance=1e-3):
        r"""
        Return a maximum packing.

        A maximum packing in a hypergraph is collection of disjoint sets/blocks
        of maximal cardinality. This problem is NP-complete in general, and in
        particular on 3-uniform hypergraphs. It is solved here with an Integer
        Linear Program.

        For more information, see the :wikipedia:`Packing_in_a_hypergraph`.

        INPUT:

        - ``solver`` -- (default: ``None``) specify a Mixed Integer Linear
          Programming (MILP) solver to be used. If set to ``None``, the default
          one is used. For more information on LP solvers and which default
          solver is used, see the method :meth:`solve
          <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
          :class:`MixedIntegerLinearProgram
          <sage.numerical.mip.MixedIntegerLinearProgram>`.

        - ``verbose`` -- integer (default: 0); sets the level of
          verbosity. Set to 0 by default, which means quiet.

        - ``integrality_tolerance`` -- parameter for use with MILP solvers over
          an inexact base ring; see
          :meth:`MixedIntegerLinearProgram.get_values`.

        EXAMPLES::

            sage: P = IncidenceStructure([[1,2],[3,4],[2,3]]).packing()                 # needs sage.numerical.mip
            sage: sorted(sorted(b) for b in P)                                          # needs sage.numerical.mip
            [[1, 2], [3, 4]]
            sage: len(designs.steiner_triple_system(9).packing())                       # needs sage.numerical.mip
            3
        """
        from sage.numerical.mip import MixedIntegerLinearProgram

        # List of blocks containing a given point x
        d = [[] for _ in self._points]
        for i, B in enumerate(self._blocks):
            for x in B:
                d[x].append(i)

        p = MixedIntegerLinearProgram(solver=solver)
        b = p.new_variable(binary=True)
        for x, L in enumerate(d):  # Set of disjoint blocks
            p.add_constraint(p.sum([b[i] for i in L]) <= 1)

        # Maximum number of blocks
        p.set_objective(p.sum([b[i] for i in range(self.num_blocks())]))

        p.solve(log=verbose)

        values = p.get_values(b, convert=bool, tolerance=integrality_tolerance)
        return [[self._points[x] for x in self._blocks[i]]
                for i, v in values.items() if v]

    def is_t_design(self, t=None, v=None, k=None, l=None, return_parameters=False):
        r"""
        Test whether ``self`` is a `t-(v,k,l)` design.

        A `t-(v,k,\lambda)` (sometimes called `t`-design for short) is a block
        design in which:

        - the underlying set has cardinality `v`
        - the blocks have size `k`
        - each `t`-subset of points is covered by `\lambda` blocks

        INPUT:

        - ``t``, ``v``, ``k``, ``l`` -- integers; their value is set to
          ``None`` by default. The function tests whether the design is a
          `t-(v,k,l)` design using the provided values and guesses the
          others. Note that ``l`` cannot be specified if ``t`` is not.

        - ``return_parameters`` -- boolean; whether to return the parameters of
          the `t`-design. If set to ``True``, the function returns a pair
          ``(boolean_answer,(t,v,k,l))``.

        EXAMPLES::

            sage: fano_blocks = [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]]
            sage: BD = IncidenceStructure(7, fano_blocks)
            sage: BD.is_t_design()
            True
            sage: BD.is_t_design(return_parameters=True)
            (True, (2, 7, 3, 1))
            sage: BD.is_t_design(2, 7, 3, 1)
            True
            sage: BD.is_t_design(1, 7, 3, 3)
            True
            sage: BD.is_t_design(0, 7, 3, 7)
            True

            sage: BD.is_t_design(0,6,3,7) or BD.is_t_design(0,7,4,7) or BD.is_t_design(0,7,3,8)
            False

            sage: BD = designs.AffineGeometryDesign(3, 1, GF(2))                        # needs sage.combinat sage.modules
            sage: BD.is_t_design(1)                                                     # needs sage.combinat sage.modules
            True
            sage: BD.is_t_design(2)                                                     # needs sage.combinat sage.modules
            True

        Steiner triple and quadruple systems are other names for `2-(v,3,1)` and
        `3-(v,4,1)` designs::

            sage: S3_9 = designs.steiner_triple_system(9)
            sage: S3_9.is_t_design(2,9,3,1)
            True

            sage: blocks = designs.steiner_quadruple_system(8)
            sage: S4_8 = IncidenceStructure(8, blocks)
            sage: S4_8.is_t_design(3,8,4,1)
            True

            sage: blocks = designs.steiner_quadruple_system(14)
            sage: S4_14 = IncidenceStructure(14, blocks)
            sage: S4_14.is_t_design(3,14,4,1)
            True

        Some examples of Witt designs that need the gap database::

            sage: # optional - gap_package_design
            sage: BD = designs.WittDesign(9)
            sage: BD.is_t_design(2,9,3,1)
            True
            sage: W12 = designs.WittDesign(12)
            sage: W12.is_t_design(5,12,6,1)
            True
            sage: W12.is_t_design(4)
            True

        Further examples::

            sage: D = IncidenceStructure(4,[[],[]])
            sage: D.is_t_design(return_parameters=True)
            (True,  (0, 4, 0, 2))

            sage: D = IncidenceStructure(4, [[0,1],[0,2],[0,3]])
            sage: D.is_t_design(return_parameters=True)
            (True, (0, 4, 2, 3))

            sage: D = IncidenceStructure(4, [[0],[1],[2],[3]])
            sage: D.is_t_design(return_parameters=True)
            (True, (1, 4, 1, 1))

            sage: D = IncidenceStructure(4,[[0,1],[2,3]])
            sage: D.is_t_design(return_parameters=True)
            (True, (1, 4, 2, 1))

            sage: D = IncidenceStructure(4, [list(range(4))])
            sage: D.is_t_design(return_parameters=True)
            (True, (4, 4, 4, 1))

        TESTS::

            sage: blocks = designs.steiner_quadruple_system(8)
            sage: S4_8 = IncidenceStructure(8, blocks)
            sage: R = list(range(15))
            sage: [(v,k,l) for v in R for k in R for l in R if S4_8.is_t_design(3,v,k,l)]
            [(8, 4, 1)]
            sage: [(v,k,l) for v in R for k in R for l in R if S4_8.is_t_design(2,v,k,l)]
            [(8, 4, 3)]
            sage: [(v,k,l) for v in R for k in R for l in R if S4_8.is_t_design(1,v,k,l)]
            [(8, 4, 7)]
            sage: [(v,k,l) for v in R for k in R for l in R if S4_8.is_t_design(0,v,k,l)]
            [(8, 4, 14)]

            sage: # needs sage.rings.finite_rings
            sage: A = designs.AffineGeometryDesign(3, 1, GF(2))
            sage: A.is_t_design(return_parameters=True)
            (True, (2, 8, 2, 1))
            sage: A = designs.AffineGeometryDesign(4, 2, GF(2))
            sage: A.is_t_design(return_parameters=True)
            (True, (3, 16, 4, 1))

            sage: I = IncidenceStructure(2, [])
            sage: I.is_t_design(return_parameters=True)
            (True, (0, 2, 0, 0))
            sage: I = IncidenceStructure(2, [[0],[0,1]])
            sage: I.is_t_design(return_parameters=True)
            (False, (0, 0, 0, 0))

        Verify that :issue:`38454` is fixed::

            sage: I = IncidenceStructure(points=[0,1,2,3,4,5],
            ....:                        blocks=[[0,1], [1,2], [0,2]])
            sage: I.is_t_design(return_parameters=True)
            (True, (0, 6, 2, 3))
        """
        from sage.arith.misc import binomial

        # Missing parameters ?
        if v is None:
            v = self.num_points()

        if k is None:
            k = len(self._blocks[0]) if self._blocks else 0

        if l is not None and t is None:
            raise ValueError("t must be set when l=None")

        b = self.num_blocks()

        # Trivial wrong answers
        if (any(len(block) != k for block in self._blocks) or  # non k-uniform
                v != self.num_points()):
            return (False, (0, 0, 0, 0)) if return_parameters else False

        # Trivial case t>k
        if (t is not None and t > k):
            if (l is None or l == 0):
                return (True, (t, v, k, 0)) if return_parameters else True
            else:
                return (False, (0, 0, 0, 0)) if return_parameters else False

        # Trivial case k=0
        if k == 0:
            if (l is None or l == 0):
                return (True, (0, v, k, b)) if return_parameters else True
            else:
                return (False, (0, 0, 0, 0)) if return_parameters else False

        # Trivial case k=v (includes v=0)
        if k == v:
            if t is None:
                t = v
            if l is None or b == l:
                return (True, (t, v, k, b)) if return_parameters else True
            else:
                return (True, (0, 0, 0, 0)) if return_parameters else False

        # Handbook of combinatorial design theorem II.4.8:
        #
        # a t-(v,k,l) is also a t'-(v,k,l')
        # for t' < t and l' = l* binomial(v-t',t-t') / binomial(k-t',t-t')
        #
        # We look for the largest t such that self is a t-design
        from itertools import combinations
        for tt in (range(1, k + 1) if t is None else [t]):
            # is lambda an integer?
            if (b * binomial(k, tt)) % binomial(v, tt):
                tt -= 1
                break

            s = {}
            for block in self._blocks:
                for i in combinations(block, tt):
                    s[i] = s.get(i, 0) + 1

            if (len(s) != binomial(v, tt)) or (len(set(s.values())) != 1):
                tt -= 1
                break

            ll = (b * binomial(k, tt)) // binomial(v, tt)

        if ((t is not None and t != tt) or
                (l is not None and l != ll)):
            return (False, (0, 0, 0, 0)) if return_parameters else False
        else:
            if tt == 0:
                ll = b
            return (True, (tt, v, k, ll)) if return_parameters else True

    def is_generalized_quadrangle(self, verbose=False, parameters=False):
        r"""
        Test if the incidence structure is a generalized quadrangle.

        An incidence structure is a generalized quadrangle iff (see [BH2012]_,
        section 9.6):

        - two blocks intersect on at most one point.

        - For every point `p` not in a block `B`, there is a unique block `B'`
          intersecting both `\{p\}` and `B`

        It is a *regular* generalized quadrangle if furthermore:

        - it is `s+1`-:meth:`uniform <is_uniform>` for some positive integer `s`.

        - it is `t+1`-:meth:`regular <is_regular>` for some positive integer `t`.

        For more information, see the :wikipedia:`Generalized_quadrangle`.

        .. NOTE::

            Some references (e.g. [PT2009]_ or
            :wikipedia:`Generalized_quadrangle`) only allow *regular*
            generalized quadrangles. To use such a definition, see the
            ``parameters`` optional argument described below, or the methods
            :meth:`is_regular` and :meth:`is_uniform`.

        INPUT:

        - ``verbose`` -- boolean; whether to print an explanation when the
          instance is not a generalized quadrangle

        - ``parameters`` -- (boolean; ``False``); if set to ``True``, the
          function returns a pair ``(s,t)`` instead of ``True`` answers. In this
          case, `s` and `t` are the integers defined above if they exist (each
          can be set to ``False`` otherwise).

        EXAMPLES::

            sage: h = designs.CremonaRichmondConfiguration()                            # needs networkx
            sage: h.is_generalized_quadrangle()                                         # needs networkx
            True

        This is actually a *regular* generalized quadrangle::

            sage: h.is_generalized_quadrangle(parameters=True)                          # needs networkx
            (2, 2)

        TESTS::

            sage: H = IncidenceStructure((2*graphs.CompleteGraph(3)).edges(sort=True, labels=False))
            sage: H.is_generalized_quadrangle(verbose=True)                             # needs sage.modules
            Some point is at distance >3 from some block.
            False

            sage: G = graphs.CycleGraph(5)
            sage: B = list(G.subgraph_search_iterator(graphs.PathGraph(3),              # needs sage.modules
            ....:                                     return_graphs=False))
            sage: H = IncidenceStructure(B)                                             # needs sage.modules
            sage: H.is_generalized_quadrangle(verbose=True)                             # needs sage.modules
            Two blocks intersect on >1 points.
            False

            sage: hypergraphs.CompleteUniform(4,2).is_generalized_quadrangle(verbose=1)             # needs sage.modules
            Some point has two projections on some line.
            False
        """
        # The distance between a point and a line in the incidence graph is odd
        # and must be <= 3. Thus, the diameter is at most 4
        g = self.incidence_graph()
        if g.diameter() > 4:
            if verbose:
                print("Some point is at distance >3 from some block.")
            return False

        # There is a unique projection of a point on a line. Thus, the girth of
        # g is at least 7
        girth = g.girth()
        if girth == 4:
            if verbose:
                print("Two blocks intersect on >1 points.")
            return False
        elif girth == 6:
            if verbose:
                print("Some point has two projections on some line.")
            return False

        if parameters:
            s = self.is_uniform()
            t = self.is_regular()
            s = s - 1 if (s is not False and s >= 2) else False
            t = t - 1 if (t is not False and t >= 2) else False
            return (s, t)
        else:
            return True

    def dual(self, algorithm=None):
        """
        Return the dual of the incidence structure.

        INPUT:

        - ``algorithm`` -- whether to use Sage's implementation
          (``algorithm=None``, default) or use GAP's (``algorithm='gap'``)

          .. NOTE::

              The ``algorithm='gap'`` option requires GAP's Design package
              (included in the ``gap_packages`` Sage spkg).

        EXAMPLES:

        The dual of a projective plane is a projective plane::

            sage: PP = designs.DesarguesianProjectivePlaneDesign(4)                     # needs sage.rings.finite_rings
            sage: PP.dual().is_t_design(return_parameters=True)                         # needs sage.modules sage.rings.finite_rings
            (True, (2, 21, 5, 1))

        TESTS::

            sage: D = IncidenceStructure(4, [[0,2],[1,2,3],[2,3]]); D
            Incidence structure with 4 points and 3 blocks
            sage: D.dual()                                                              # needs sage.modules
            Incidence structure with 3 points and 4 blocks
            sage: print(D.dual(algorithm='gap'))            # optional - gap_package_design
            Incidence structure with 3 points and 4 blocks
            sage: blocks = [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]]
            sage: BD = IncidenceStructure(7, blocks, name='FanoPlane'); BD
            Incidence structure with 7 points and 7 blocks
            sage: print(BD.dual(algorithm='gap'))           # optional - gap_package_design
            Incidence structure with 7 points and 7 blocks
            sage: BD.dual()                                                             # needs sage.modules
            Incidence structure with 7 points and 7 blocks

        REFERENCE:

        - Leonard Soicher, :gap_package:`Design package manual <design/htm/CHAP003.htm>`
        """
        if algorithm == "gap":
            libgap.load_package("design")
            DD = libgap(self).DualBlockDesign()
            v = DD['v'].sage()
            gB = [[x - 1 for x in b] for b in DD['blocks'].sage()]
            return IncidenceStructure(list(range(v)), gB, name=None, check=False)

        return IncidenceStructure(
            incidence_matrix=self.incidence_matrix().transpose(),
            check=False)

    def automorphism_group(self):
        r"""
        Return the subgroup of the automorphism group of the incidence graph
        which respects the P B partition. It is (isomorphic to) the automorphism
        group of the block design, although the degrees differ.

        EXAMPLES::

            sage: # needs sage.groups sage.rings.finite_rings
            sage: P = designs.DesarguesianProjectivePlaneDesign(2); P
            (7,3,1)-Balanced Incomplete Block Design
            sage: G = P.automorphism_group()
            sage: G.is_isomorphic(PGL(3,2))
            True
            sage: G
            Permutation Group with generators [...]
            sage: G.cardinality()
            168

        A non self-dual example::

            sage: IS = IncidenceStructure(list(range(4)), [[0,1,2,3],[1,2,3]])
            sage: IS.automorphism_group().cardinality()                                 # needs sage.groups
            6
            sage: IS.dual().automorphism_group().cardinality()                          # needs sage.groups sage.modules
            1

        Examples with non-integer points::

            sage: I = IncidenceStructure('abc', ('ab','ac','bc'))
            sage: I.automorphism_group()                                                # needs sage.groups
            Permutation Group with generators [('b','c'), ('a','b')]
            sage: IncidenceStructure([[(1,2),(3,4)]]).automorphism_group()              # needs sage.groups
            Permutation Group with generators [((1,2),(3,4))]
        """
        from sage.graphs.graph import Graph
        from sage.groups.perm_gps.permgroup import PermutationGroup
        g = Graph()
        n = self.num_points()
        g.add_edges((i + n, x) for i, b in enumerate(self._blocks) for x in b)
        ag = g.automorphism_group(partition=[list(range(n)),
                                             list(range(n, n + self.num_blocks()))])

        if self._point_to_index:
            gens = [[tuple([self._points[i] for i in cycle if (not cycle or cycle[0] < n)])
                     for cycle in g.cycle_tuples()]
                    for g in ag.gens()]
        else:
            gens = [[tuple(cycle) for cycle in g.cycle_tuples() if (not cycle or cycle[0] < n)]
                    for g in ag.gens()]

        return PermutationGroup(gens, domain=self._points)

    def is_resolvable(self, certificate=False, solver=None, verbose=0, check=True,
                      *, integrality_tolerance=1e-3):
        r"""
        Test whether the hypergraph is resolvable.

        A hypergraph is said to be resolvable if its sets can be partitionned
        into classes, each of which is a partition of the ground set.

        .. NOTE::

            This problem is solved using an Integer Linear Program, and GLPK
            (the default LP solver) has been reported to be very slow on some
            instances. If you hit this wall, consider installing a more powerful
            MILP solver (CPLEX, Gurobi, ...).

        INPUT:

        - ``certificate`` -- boolean; whether to return the classes along with
          the binary answer (see examples below)

        - ``solver`` -- (default: ``None``) specify a Mixed Integer Linear
          Programming (MILP) solver to be used. If set to ``None``, the default
          one is used. For more information on MILP solvers and which default
          solver is used, see the method :meth:`solve
          <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
          :class:`MixedIntegerLinearProgram
          <sage.numerical.mip.MixedIntegerLinearProgram>`.

        - ``verbose`` -- integer (default: 0); sets the level of
          verbosity. Set to 0 by default, which means quiet.

        - ``check`` -- boolean (default: ``True``); whether to check that
          output is correct before returning it. As this is expected to be
          useless, you may want to disable it whenever you want speed.

        - ``integrality_tolerance`` -- parameter for use with MILP solvers over
          an inexact base ring; see :meth:`MixedIntegerLinearProgram.get_values`

        EXAMPLES:

        Some resolvable designs::

            sage: TD = designs.transversal_design(2,2,resolvable=True)
            sage: TD.is_resolvable()
            True

            sage: AG = designs.AffineGeometryDesign(3,1,GF(2))                          # needs sage.combinat sage.modules
            sage: AG.is_resolvable()                                                    # needs sage.combinat sage.modules
            True

        Their classes::

            sage: b, cls = TD.is_resolvable(True)
            sage: b
            True
            sage: cls # random
            [[[0, 3], [1, 2]], [[1, 3], [0, 2]]]

            sage: # needs sage.combinat
            sage: b, cls = AG.is_resolvable(True)
            sage: b
            True
            sage: cls  # random
            [[[6, 7], [4, 5], [0, 1], [2, 3]],
             [[5, 7], [0, 4], [3, 6], [1, 2]],
             [[0, 2], [4, 7], [1, 3], [5, 6]],
             [[3, 4], [0, 7], [1, 5], [2, 6]],
             [[3, 7], [1, 6], [0, 5], [2, 4]],
             [[0, 6], [2, 7], [1, 4], [3, 5]],
             [[4, 6], [0, 3], [2, 5], [1, 7]]]

        A non-resolvable design::

            sage: Fano = designs.balanced_incomplete_block_design(7,3)                  # needs sage.schemes
            sage: Fano.is_resolvable()                                                  # needs sage.schemes
            False
            sage: Fano.is_resolvable(True)                                              # needs sage.schemes
            (False, [])

        TESTS::

            sage: # needs sage.combinat
            sage: _, cls1 = AG.is_resolvable(certificate=True)
            sage: _, cls2 = AG.is_resolvable(certificate=True)
            sage: cls1 is cls2
            False
        """
        if self._classes is None:
            degrees = set(self.degrees().values())
            if len(degrees) != 1:
                self._classes = False
            else:
                from sage.numerical.mip import MixedIntegerLinearProgram
                from sage.numerical.mip import MIPSolverException
                n_classes = degrees.pop()
                p = MixedIntegerLinearProgram(solver=solver)
                b = p.new_variable(binary=True)
                domain = list(range(self.num_points()))

                # Lists of blocks containing i for every i
                dual = [[] for _ in domain]
                for i, B in enumerate(self._blocks):
                    for x in B:
                        dual[x].append(i)

                # Each class is a partition
                for t in range(n_classes):
                    for x in domain:
                        p.add_constraint(p.sum(b[t, i] for i in dual[x]) == 1)

                # Each set appears exactly once
                for i in range(len(self._blocks)):
                    p.add_constraint(p.sum(b[t, i] for t in range(n_classes)) == 1)

                try:
                    p.solve(log=verbose)
                except MIPSolverException:
                    self._classes = False
                else:
                    # each class is stored as the list of indices of its blocks
                    self._classes = [[] for _ in range(n_classes)]
                    for (t, i), v in p.get_values(b, convert=bool, tolerance=integrality_tolerance).items():
                        if v:
                            self._classes[t].append(self._blocks[i])

        if check and self._classes is not False:
            assert sorted(id(c) for cls in self._classes for c in cls) == sorted(id(b) for b in self._blocks), "some set does not appear exactly once"
            domain = list(range(self.num_points()))
            for i, c in enumerate(self._classes):
                assert sorted(sum(c, [])) == domain, "class {} is not a partition".format(i)

        if self._classes is False:
            return (False, []) if certificate else False

        if certificate:
            if self._point_to_index is None:
                classes = [[block[:] for block in classs] for classs in self._classes]
            else:
                classes = [[[self._points[i] for i in block] for block in classs] for classs in self._classes]

            return (True, classes)

        else:
            return True

    def coloring(self, k=None, solver=None, verbose=0,
                 *, integrality_tolerance=1e-3):
        r"""
        Compute a (weak) `k`-coloring of the hypergraph.

        A weak coloring of a hypergraph `\mathcal H` is an assignment of colors
        to its vertices such that no set is monochromatic.

        INPUT:

        - ``k`` -- integer; compute a coloring with `k` colors if an integer is
          provided, otherwise returns an optimal coloring (i.e. with the minimum
          possible number of colors).

        - ``solver`` -- (default: ``None``) specify a Mixed Integer Linear
          Programming (MILP) solver to be used. If set to ``None``, the default
          one is used. For more information on MILP solvers and which default
          solver is used, see the method :meth:`solve
          <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
          :class:`MixedIntegerLinearProgram
          <sage.numerical.mip.MixedIntegerLinearProgram>`.

        - ``verbose`` -- nonnegative integer (default: `0`); set the level
          of verbosity you want from the linear program solver. Since the
          problem is `NP`-complete, its solving may take some time depending on
          the graph. A value of `0` means that there will be no message printed by
          the solver.

        - ``integrality_tolerance`` -- parameter for use with MILP solvers over
          an inexact base ring; see :meth:`MixedIntegerLinearProgram.get_values`

        EXAMPLES:

        The Fano plane has chromatic number 3::

            sage: len(designs.steiner_triple_system(7).coloring())                      # needs sage.numerical.mip
            3

        One admissible 3-coloring::

            sage: designs.steiner_triple_system(7).coloring()   # not tested            # needs sage.numerical.mip
            [[0, 2, 5, 1], [4, 3], [6]]

        The chromatic number of a graph is equal to the chromatic number of its
        2-uniform corresponding hypergraph::

            sage: g = graphs.PetersenGraph()
            sage: H = IncidenceStructure(g.edges(sort=True, labels=False))
            sage: len(g.coloring())
            3
            sage: len(H.coloring())                                                     # needs sage.numerical.mip
            3
        """
        if k is None:
            for k in range(self.num_points() + 1):
                try:
                    return self.coloring(k)
                except ValueError:
                    pass

        if k == 0:
            if self.num_points():
                raise ValueError("Only empty hypergraphs are 0-chromatic")
            return []
        elif any(len(x) == 1 for x in self._blocks):
            raise RuntimeError("No coloring can be defined "
                               "when there is a set of size 1")
        elif k == 1:
            if any(self._blocks):
                raise ValueError("This hypergraph contains a set. "
                                 "It is not 1-chromatic")
            return [self.ground_set()]

        from sage.numerical.mip import MixedIntegerLinearProgram, MIPSolverException
        p = MixedIntegerLinearProgram(solver=solver)
        b = p.new_variable(binary=True)

        for x in range(self.num_points()):
            p.add_constraint(p.sum(b[x, i] for i in range(k)) == 1)

        for s in self._blocks:
            for i in range(k):
                p.add_constraint(p.sum(b[x, i] for x in s) <= len(s) - 1)

        try:
            p.solve(log=verbose)
        except MIPSolverException:
            raise ValueError("This hypergraph is not {}-colorable".format(k))

        col = [[] for _ in range(k)]

        for (x, i), v in p.get_values(b, convert=bool, tolerance=integrality_tolerance).items():
            if v:
                col[i].append(self._points[x])

        return col

    def edge_coloring(self) -> list:
        r"""
        Compute a proper edge-coloring.

        A proper edge-coloring is an assignment of colors to the sets of the
        incidence structure such that two sets with non-empty intersection
        receive different colors. The coloring returned minimizes the number of
        colors.

        OUTPUT: a partition of the sets into color classes

        EXAMPLES::

            sage: # needs cliquer
            sage: H = Hypergraph([{1,2,3},{2,3,4},{3,4,5},{4,5,6}]); H
            Incidence structure with 6 points and 4 blocks
            sage: C = H.edge_coloring()
            sage: C # random
            [[[3, 4, 5]], [[2, 3, 4]], [[4, 5, 6], [1, 2, 3]]]
            sage: Set(map(Set,sum(C,[]))) == Set(map(Set,H.blocks()))
            True
        """
        from sage.graphs.graph import Graph
        blocks = self.blocks()
        blocks_sets = [frozenset(b) for b in blocks]
        g = Graph([list(range(self.num_blocks())),
                   lambda x, y: len(blocks_sets[x] & blocks_sets[y])],
                  loops=False)
        return [[blocks[i] for i in C] for C in g.coloring(algorithm='MILP')]

    def _spring_layout(self):
        r"""
        Return a spring layout for the points.

        The layout is computed by creating a graph `G` on the points *and* sets
        of the incidence structure. Each set is then made adjacent in `G` with
        all points it contains before a spring layout is computed for this
        graph. The position of the points in the graph gives the position of the
        points in the final drawing.

        .. NOTE::

            This method also returns the position of the "fake" points,
            i.e. those representing the sets.

        EXAMPLES::

            sage: # needs sage.plot
            sage: H = Hypergraph([{1,2,3},{2,3,4},{3,4,5},{4,5,6}]); H
            Incidence structure with 6 points and 4 blocks
            sage: L = H._spring_layout()
            sage: L  # random
            {1: (0.238, -0.926),
             2: (0.672, -0.518),
             3: (0.449, -0.225),
             4: (0.782, 0.225),
             5: (0.558, 0.518),
             6: (0.992, 0.926),
             {3, 4, 5}: (0.504, 0.173),
             {2, 3, 4}: (0.727, -0.173),
             {4, 5, 6}: (0.838, 0.617),
             {1, 2, 3}: (0.393, -0.617)}
            sage: all(v in L for v in H.ground_set())
            True
            sage: all(v in L for v in map(Set, H.blocks()))
            True
        """
        from sage.graphs.graph import Graph

        g = Graph()
        for s in map(Set, self.blocks()):
            for x in s:
                g.add_edge((0, s), (1, x))

        _ = g.plot(iterations=50000, save_pos=True)

        # The values are rounded as TikZ does not like accuracy.
        return {k[1]: (round(x, 3), round(y, 3))
                for k, (x, y) in g.get_pos().items()}

    def _latex_(self) -> str:
        r"""
        Return a TikZ representation of the incidence structure.

        EXAMPLES::

            sage: H = Hypergraph([{1,2,3},{2,3,4},{3,4,5},{4,5,6}]); H
            Incidence structure with 6 points and 4 blocks
            sage: view(H) # not tested

        With sets of size 4::

            sage: g = graphs.Grid2dGraph(5,5)
            sage: C4 = graphs.CycleGraph(4)
            sage: sets = Set(map(Set, g.subgraph_search_iterator(C4,                    # needs sage.modules
            ....:                                                return_graphs=False)))
            sage: H = Hypergraph(sets)                                                  # needs sage.modules
            sage: view(H)                       # not tested                            # needs sage.modules sage.plot

        TESTS::

            # verify that :issue:`30976` is fixed
            sage: IS = IncidenceStructure([1,2,3], [[1,2], [2,3]])
            sage: if latex.has_file("tikz.sty"):          # optional - latex
            ....:     IS._latex_()
            ...UserWarning:
            The hypergraph is drawn as a set of closed curves...
            \begin{tikzpicture}...
            \draw... -- ...;
            \draw... -- ...;
             \draw node...;
             \draw node...;
             \draw node...;
            \end{tikzpicture}
        """
        from sage.functions.trig import arctan2

        from warnings import warn
        warn("\nThe hypergraph is drawn as a set of closed curves. The curve "
             "representing a set S goes **THROUGH** the points contained "
             "in S.\n A point which is encircled by a curve but is not located "
             "on its boundary is **NOT** included in the corresponding set.\n"
             "\n"
             "The colors are picked for readability and have no other meaning.")

        latex.add_package_to_preamble_if_available("tikz")

        if not latex.has_file("tikz.sty"):
            raise RuntimeError("You must have TikZ installed in order "
                               "to draw a hypergraph.")

        domain = self.ground_set()
        pos = self._spring_layout()
        tex = "\\begin{tikzpicture}[scale=3]\n"

        colors = ["black", "red", "green", "blue", "cyan",
                  "magenta", "yellow", "pink", "brown"]
        colored_sets = [(s, i) for i, S in enumerate(self.edge_coloring()) for s in S]

        # Prints each set with its color
        for s, i in colored_sets:
            current_color = colors[i % len(colors)]

            if len(s) == 2:
                s = list(s)
                tex += ("\\draw[color="+str(current_color)+"," +
                        "line width=.1cm,opacity = .6] " +
                        str(pos[s[0]])+" -- "+str(pos[s[1]])+";\n")
                continue

            tex += ("\\draw[color="+str(current_color)+","
                    "line width=.1cm,opacity = .6,"
                    "line cap=round,"
                    "line join=round]"
                    "plot [smooth cycle,tension=1] coordinates {")

            # Reorders the vertices of s according to their angle with the
            # "center", i.e. the vertex representing the set s
            cx, cy = pos[Set(s)]
            s = [pos[_] for _ in s]
            s = sorted(s, key=lambda x_y: arctan2(x_y[0] - cx, x_y[1] - cy))

            for x in s:
                tex += str(x)+" "
            tex += "};\n"

        # Prints each vertex
        for v in domain:
            tex += "\\draw node[fill,circle,scale=.5,label={90:$"+latex(v)+"$}] at "+str(pos[v])+" {};\n"

        tex += "\\end{tikzpicture}"
        return tex

    def is_spread(self, spread) -> bool:
        r"""
        Check whether the input is a spread for ``self``.

        A spread of an incidence structure `(P, B)` is a subset of `B` which
        forms a partition of `P`.

        INPUT:

        - ``spread`` -- iterable; defines the spread

        EXAMPLES::

            sage: E = IncidenceStructure([[1, 2, 3], [4, 5, 6], [1, 5, 6]])
            sage: E.is_spread([[1, 2, 3], [4, 5, 6]])
            True
            sage: E.is_spread([1, 2, 3, 4, 5, 6])
            Traceback (most recent call last):
            ...
            TypeError: 'sage.rings.integer.Integer' object is not iterable
            sage: E.is_spread([[1, 2, 3, 4], [5, 6]])
            False

        Order of blocks or of points within each block doesn't matter::

            sage: E = IncidenceStructure([[1, 2, 3], [4, 5, 6], [1, 5, 6]])
            sage: E.is_spread([[5, 6, 4], [3, 1, 2]])
            True

        TESTS::

            sage: E = IncidenceStructure([])
            sage: E.is_spread([])
            True
            sage: E = IncidenceStructure([[1]])
            sage: E.is_spread([])
            False
            sage: E.is_spread([[1]])
            True
            sage: E = IncidenceStructure([[1], [1]])
            sage: E.is_spread([[1]])
            True
        """

        points = set(self.ground_set())
        allBlocks = set(map(frozenset, self.blocks()))
        for block in spread:
            sblock = set(block)

            if sblock not in allBlocks:
                return False

            if not points.issuperset(sblock):
                return False

            points.difference_update(sblock)

        return not points


from sage.misc.rest_index_of_methods import gen_rest_table_index
__doc__ = __doc__.format(METHODS_OF_IncidenceStructure=gen_rest_table_index(IncidenceStructure))



# sage_setup: distribution = sagemath-graphs
from sage.combinat.all__sagemath_categories import *

from sage.misc.lazy_import import lazy_import

# Posets
from sage.combinat.posets.all import *

# Trees and Tamari interval posets
from sage.combinat.ordered_tree import (OrderedTree, OrderedTrees,
                                        LabelledOrderedTree, LabelledOrderedTrees)
from sage.combinat.binary_tree import (BinaryTree, BinaryTrees,
                                       LabelledBinaryTree, LabelledBinaryTrees)
lazy_import('sage.combinat.rooted_tree', ('RootedTree', 'RootedTrees',
                                          'LabelledRootedTree', 'LabelledRootedTrees'))
lazy_import('sage.combinat.interval_posets', [
            'TamariIntervalPoset', 'TamariIntervalPosets'])

from sage.combinat.graph_path import GraphPaths

from sage.combinat.yang_baxter_graph import YangBaxterGraph

# block designs, etc
from sage.combinat.designs.all import *

# Cluster Algebras and Quivers
from sage.combinat.cluster_algebra_quiver.all import *

# Finite State Machines (Automaton, Transducer)
lazy_import('sage.combinat.finite_state_machine',
            ['Automaton', 'Transducer', 'FiniteStateMachine'])
lazy_import('sage.combinat.finite_state_machine_generators',
            ['automata', 'transducers'])
del lazy_import



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
r"""
Some fast computations for finite posets
"""
# ****************************************************************************
#       Copyright (C) 2020 Frédéric Chapoton <chapoton@unistra.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.misc.lazy_import import LazyImport
from sage.sets.recursively_enumerated_set import RecursivelyEnumeratedSet_forest

coxeter_matrix_fast = LazyImport('sage.combinat.posets.hasse_cython_flint', 'coxeter_matrix_fast',
                                 deprecation=35741)
moebius_matrix_fast = LazyImport('sage.combinat.posets.hasse_cython_flint', 'moebius_matrix_fast',
                                 deprecation=35741)


class IncreasingChains(RecursivelyEnumeratedSet_forest):
    r"""
    The enumerated set of increasing chains.

    INPUT:

    - ``positions`` -- list of sets of integers describing the poset,
      as given by the lazy attribute ``_leq_storage`` of Hasse diagrams

    - ``element_constructor`` -- used to determine the type of chains,
      for example :class:`list` or :class:`tuple`

    - ``exclude`` -- list of integers that should not belong to the chains

    - ``conversion`` -- (optional) list of elements of the poset

    If ``conversion`` is provided, it is used to convert chain elements
    to elements of this list.

    EXAMPLES::

        sage: from sage.combinat.posets.hasse_cython import IncreasingChains
        sage: D = IncreasingChains([{0,1},{1}], list, []); D
        An enumerated set with a forest structure
        sage: D.cardinality()
        4
        sage: list(D)
        [[], [0], [0, 1], [1]]
    """
    def __init__(self, list positions, element_constructor,
                 list exclude, conversion=None):
        """
        The enumerated set of increasing chains.

        TESTS::

            sage: from sage.combinat.posets.hasse_cython import IncreasingChains
            sage: D = IncreasingChains([{0,1},{1}], list, [])
            sage: TestSuite(D).run(skip='_test_pickling')
        """
        cdef int i
        cdef Py_ssize_t n = len(positions)
        self._n = n
        if exclude is not None:
            self._greater_than = [gti.difference(exclude) for gti in positions]
            self._vertices = [u for u in range(n) if u not in exclude]
        else:
            self._greater_than = positions
            self._vertices = list(range(n))

        self._constructor = element_constructor
        self._conversion = conversion
        if conversion is not None:
            self._from_poset = {elt: i for i, elt in enumerate(conversion)}

        self._roots = (tuple(),)

        RecursivelyEnumeratedSet_forest.__init__(self, algorithm='depth',
                    category=FiniteEnumeratedSets())

    def __contains__(self, tup):
        """
        Membership testing.

        If ``conversion`` was provided, it first converts elements of ``tup``
        to integers.

        EXAMPLES::

            sage: from sage.combinat.posets.hasse_cython import IncreasingChains
            sage: D = IncreasingChains([{0,1},{1}], list, [])
            sage: all(x in D for x in D)
            True
            sage: [2] in D
            False
            sage: [1,1] in D
            False

            sage: P = Poset({'a':['b'],'b':[]})
            sage: ['a'] in P.chains()
            True

        TESTS::

            sage: from sage.combinat.posets.hasse_cython import IncreasingChains
            sage: D = IncreasingChains([{0,1},{1}], list, [])
            sage: all(tuple(x) in D for x in D)
            True
        """
        cdef int k
        cdef Py_ssize_t i, x, y
        if not tup:
            return True
        if self._conversion is not None:
            tup = [self._from_poset[elt] for elt in tup]
        if any(not(0 <= i < self._n) for i in tup):
            return False
        y = tup[0]
        for k in range(1, len(tup)):
            x = y
            y = tup[k]
            if x == y or y not in self._greater_than[x]:
                return False
        return True

    def post_process(self, chain):
        """
        Create a chain from the internal object.

        If ``conversion`` was provided, it first converts elements of the
        chain to elements of this list.

        Then the given ``element_constructor`` is applied to the chain.

        EXAMPLES::

            sage: from sage.combinat.posets.hasse_cython import IncreasingChains
            sage: D = IncreasingChains([{0,1},{1}], list, [])
            sage: D.post_process((0,1))
            [0, 1]

            sage: P = Poset({'a':['b'],'b':[]})
            sage: list(P.chains())
            [[], ['a'], ['a', 'b'], ['b']]
        """
        cdef Py_ssize_t i
        if self._conversion is not None:
            return self._constructor(self._conversion[i] for i in chain)
        return self._constructor(chain)

    def children(self, chain):
        """
        Return the children of a chain, by adding one largest element.

        EXAMPLES::

            sage: from sage.combinat.posets.hasse_cython import IncreasingChains
            sage: D = IncreasingChains([{0,1},{1}], list, [])
            sage: D.children((0,))
            [(0, 1)]

            sage: P = Poset({'a':['b'],'b':[]})
            sage: next(iter(P.chains()))
            []
        """
        cdef Py_ssize_t x, y
        if not chain:
            return [(x,) for x in self._vertices]
        x = chain[-1]
        return [chain + (y,) for y in self._greater_than[x] if x != y]



# sage_setup: distribution = sagemath-graphs
r"""
Tutte polynomial

This module implements a deletion-contraction algorithm for computing
the Tutte polynomial as described in the paper [HPR2010]_.

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :func:`tutte_polynomial` | Compute the Tutte polynomial of the input graph

Authors:

- Mike Hansen (06-2013), Implemented the algorithm.
- Jernej Azarija (06-2013), Tweaked the code, added documentation

Definition
-----------

Given a graph `G`, with `n` vertices and `m` edges and `k(G)`
connected components we define the Tutte polynomial of `G` as

.. MATH::

    \sum_H (x-1) ^{k(H) - c} (y-1)^{k(H) - |E(H)|-n}

where the sum ranges over all induced subgraphs `H` of `G`.

Functions
---------
"""

from contextlib import contextmanager
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.misc_c import prod
from sage.rings.integer_ring import ZZ
from sage.misc.decorators import sage_wraps

######################
# Graph Modification #
######################


@contextmanager
def removed_multiedge(G, unlabeled_edge):
    r"""
    A context manager which removes an edge with multiplicity from the
    graph `G` and restores it upon exiting.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import removed_multiedge
        sage: G = Graph(multiedges=True)
        sage: G.add_edges([(0,1,'a'),(0,1,'b')])
        sage: G.edges(sort=True)
        [(0, 1, 'a'), (0, 1, 'b')]
        sage: with removed_multiedge(G,(0,1)) as Y:
        ....:     G.edges(sort=True)
        []
        sage: G.edges(sort=True)
        [(0, 1, 'a'), (0, 1, 'b')]
    """
    u, v = unlabeled_edge
    edges = G.edge_boundary([u], [v], labels=True)
    G.delete_multiedge(u, v)
    try:
        yield
    finally:
        G.add_edges(edges)


@contextmanager
def removed_edge(G, edge):
    r"""
    A context manager which removes an edge from the graph `G` and
    restores it upon exiting.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import removed_edge
        sage: G = Graph()
        sage: G.add_edge(0,1)
        sage: G.edges(sort=True)
        [(0, 1, None)]
        sage: with removed_edge(G,(0,1)) as Y:
        ....:     G.edges(sort=True); G.vertices(sort=True)
        []
        [0, 1]
        sage: G.edges(sort=True)
        [(0, 1, None)]
    """
    G.delete_edge(edge)
    try:
        yield
    finally:
        G.add_edge(edge)


@contextmanager
def contracted_edge(G, unlabeled_edge):
    r"""
    Delete the first vertex in the edge, and make all the edges that
    went from it go to the second vertex.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import contracted_edge
        sage: G = Graph(multiedges=True)
        sage: G.add_edges([(0,1,'a'),(1,2,'b'),(0,3,'c')])
        sage: G.edges(sort=True)
        [(0, 1, 'a'), (0, 3, 'c'), (1, 2, 'b')]
        sage: with contracted_edge(G,(0,1)) as Y:
        ....:     G.edges(sort=True); G.vertices(sort=True)
        [(1, 2, 'b'), (1, 3, 'c')]
        [1, 2, 3]
        sage: G.edges(sort=True)
        [(0, 1, 'a'), (0, 3, 'c'), (1, 2, 'b')]
    """
    v1, v2 = unlabeled_edge
    loops = G.allows_loops()

    v1_edges = G.edges_incident(v1)
    G.delete_vertex(v1)
    added_edges = []

    for start, end, label in v1_edges:
        other_vertex = start if start != v1 else end
        edge = (other_vertex, v2, label)
        if loops or other_vertex != v2:
            G.add_edge(edge)
        added_edges.append(edge)

    try:
        yield
    finally:
        for edge in added_edges:
            G.delete_edge(edge)
        for edge in v1_edges:
            G.add_edge(edge)


@contextmanager
def removed_loops(G):
    r"""
    A context manager which removes all the loops in the graph `G`.
    It yields a list of the loops, and restores the loops upon
    exiting.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import removed_loops
        sage: G = Graph(multiedges=True, loops=True)
        sage: G.add_edges([(0,1,'a'),(1,2,'b'),(0,0,'c')])
        sage: G.edges(sort=True)
        [(0, 0, 'c'), (0, 1, 'a'), (1, 2, 'b')]
        sage: with removed_loops(G) as Y:
        ....:     G.edges(sort=True); G.vertices(sort=True); Y
        [(0, 1, 'a'), (1, 2, 'b')]
        [0, 1, 2]
        [(0, 0, 'c')]
        sage: G.edges(sort=True)
        [(0, 0, 'c'), (0, 1, 'a'), (1, 2, 'b')]
    """
    loops = G.loops()
    for edge in loops:
        G.delete_edge(edge)
    try:
        yield loops
    finally:
        for edge in loops:
            G.add_edge(edge)


def underlying_graph(G):
    r"""
    Given a graph `G` with multi-edges, returns a graph where all the
    multi-edges are replaced with a single edge.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import underlying_graph
        sage: G = Graph(multiedges=True)
        sage: G.add_edges([(0,1,'a'),(0,1,'b')])
        sage: G.edges(sort=True)
        [(0, 1, 'a'), (0, 1, 'b')]
        sage: underlying_graph(G).edges(sort=True)
        [(0, 1, None)]
    """
    from sage.graphs.graph import Graph
    g = Graph()
    g.allow_loops(True)
    for edge in set(G.edges(sort=False, labels=False)):
        g.add_edge(edge)
    return g


def edge_multiplicities(G):
    r"""
    Return the dictionary of multiplicities of the edges in the
    graph `G`.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import edge_multiplicities
        sage: G = Graph({1: [2,2,3], 2: [2], 3: [4,4], 4: [2,2,2]})
        sage: sorted(edge_multiplicities(G).items())
        [((1, 2), 2), ((1, 3), 1), ((2, 2), 1), ((2, 4), 3), ((3, 4), 2)]
    """
    d = {}
    for edge in G.edges(sort=False, labels=False):
        d[edge] = d.setdefault(edge, 0) + 1
    return d

########
# Ears #
########


class Ear:
    r"""
    An ear is a sequence of vertices

    Here is the definition from [HPR2010]_:

    An ear in a graph is a path `v_1 - v_2 - \dots - v_n - v_{n+1}`
    where `d(v_1) > 2`, `d(v_{n+1}) > 2` and
    `d(v_2) = d(v_3) = \dots = d(v_n) = 2`.

    A cycle is viewed as a special ear where `v_1 = v_{n+1}` and the
    restriction on the degree of this vertex is lifted.

    INPUT:
    """
    def __init__(self, graph, end_points, interior, is_cycle):
        """
        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear(G,[0,3],[1,2],False)
        """
        self.end_points = end_points
        self.interior = interior
        self.is_cycle = is_cycle
        self.graph = graph

    @property
    def s(self):
        """
        Return the number of distinct edges in this ear.

        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear(G,[0,3],[1,2],False)
            sage: E.s
            3
        """
        return len(self.interior) + 1

    @property
    def vertices(self):
        """
        Return the vertices of this ear.

        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear(G,[0,3],[1,2],False)
            sage: E.vertices
            [0, 1, 2, 3]
        """
        return sorted(self.end_points + self.interior)

    @lazy_attribute
    def unlabeled_edges(self):
        """
        Return the edges in this ear.

        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear(G,[0,3],[1,2],False)
            sage: E.unlabeled_edges
            [(0, 1), (1, 2), (2, 3)]
        """
        return self.graph.edges_incident(vertices=self.interior, labels=False)

    @staticmethod
    def find_ear(g):
        """
        Finds the first ear in a graph.

        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear.find_ear(G)
            sage: E.s
            3
            sage: E.unlabeled_edges
            [(0, 1), (1, 2), (2, 3)]
            sage: E.vertices
            [0, 1, 2, 3]
        """
        degree_two_vertices = [v for v, degree
                               in g.degree_iterator(labels=True)
                               if degree == 2]
        subgraph = g.subgraph(degree_two_vertices)
        for component in subgraph.connected_components(sort=False):
            edges = g.edges_incident(vertices=component, labels=True)
            all_vertices = sorted(set(sum([e[:2] for e in edges], ())))
            if len(all_vertices) < 3:
                continue
            end_points = [v for v in all_vertices if v not in component]
            if not end_points:
                end_points = [component[0]]

            ear_is_cycle = end_points[0] == end_points[-1]

            if ear_is_cycle:
                for e in end_points:
                    if e in component:
                        component.remove(e)

            return Ear(g, end_points, component, ear_is_cycle)

    @contextmanager
    def removed_from(self, G):
        r"""
        A context manager which removes the ear from the graph `G`.

        EXAMPLES::

            sage: G = graphs.PathGraph(4)
            sage: G.add_edges([(0,4),(0,5),(3,6),(3,7)])
            sage: len(G.edges(sort=True))
            7
            sage: from sage.graphs.tutte_polynomial import Ear
            sage: E = Ear.find_ear(G)
            sage: with E.removed_from(G) as Y:
            ....:     G.edges(sort=True)
            [(0, 4, None), (0, 5, None), (3, 6, None), (3, 7, None)]
            sage: len(G.edges(sort=True))
            7
        """
        deleted_edges = []
        for edge in G.edges_incident(vertices=self.interior, labels=True):
            G.delete_edge(edge)
            deleted_edges.append(edge)
        for v in self.interior:
            G.delete_vertex(v)

        try:
            yield
        finally:
            for edge in deleted_edges:
                G.add_edge(edge)

##################
# Edge Selection #
##################


class EdgeSelection:
    pass


class VertexOrder(EdgeSelection):
    def __init__(self, order):
        """
        EXAMPLES::

            sage: from sage.graphs.tutte_polynomial import VertexOrder
            sage: A = VertexOrder([4,6,3,2,1,7])
            sage: A.order
            [4, 6, 3, 2, 1, 7]
            sage: A.inverse_order
            {1: 4, 2: 3, 3: 2, 4: 0, 6: 1, 7: 5}
        """
        self.order = list(order)
        self.inverse_order = dict([reversed(_) for _ in enumerate(order)])

    def __call__(self, graph):
        """
        EXAMPLES::

            sage: from sage.graphs.tutte_polynomial import VertexOrder
            sage: A = VertexOrder([4,0,3,2,1,5])
            sage: G = graphs.PathGraph(6)
            sage: A(G)
            (3, 4, None)
        """
        for v in self.order:
            edges = graph.edges_incident([v])
            if edges:
                edges.sort(key=lambda x: self.inverse_order[x[0] if x[0] != v else x[1]])
                return edges[0]
        raise RuntimeError("no edges left to select")


class MinimizeSingleDegree(EdgeSelection):
    def __call__(self, graph):
        """
        EXAMPLES::

            sage: from sage.graphs.tutte_polynomial import MinimizeSingleDegree
            sage: G = graphs.PathGraph(6)
            sage: MinimizeSingleDegree()(G)
            (0, 1, None)
        """
        degrees = list(graph.degree_iterator(labels=True))
        degrees.sort(key=lambda x: x[1])  # Sort by degree
        for v, degree in degrees:
            for e in graph.edges_incident([v], labels=True):
                return e
        raise RuntimeError("no edges left to select")


class MinimizeDegree(EdgeSelection):
    def __call__(self, graph):
        """
        EXAMPLES::

            sage: from sage.graphs.tutte_polynomial import MinimizeDegree
            sage: G = graphs.PathGraph(6)
            sage: MinimizeDegree()(G)
            (0, 1, None)
        """
        degrees = dict(graph.degree_iterator(labels=True))
        edges = graph.edges(labels=True, sort=False)
        if edges:
            return min(edges, key=lambda x: degrees[x[0]] + degrees[x[1]])
        raise RuntimeError("no edges left to select")


class MaximizeDegree(EdgeSelection):
    def __call__(self, graph):
        """
        EXAMPLES::

            sage: from sage.graphs.tutte_polynomial import MaximizeDegree
            sage: G = graphs.PathGraph(6)
            sage: MaximizeDegree()(G)
            (1, 2, None)
        """
        degrees = dict(graph.degree_iterator(labels=True))
        edges = graph.edges(labels=True, sort=False)
        if edges:
            return max(edges, key=lambda x: degrees[x[0]] + degrees[x[1]])
        raise RuntimeError("no edges left to select")


###########
# Caching #
###########


def _cache_key(G):
    """
    Return the key used to cache the result for the graph G.

    This is used by the decorator :func:`_cached`.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import _cache_key
        sage: G = graphs.DiamondGraph()
        sage: print(_cache_key(G))
        ((0, 2), (0, 3), (1, 2), (1, 3), (2, 3))
    """
    return tuple(G.canonical_label().edges(labels=False, sort=True))


def _cached(func):
    """
    Wrapper used to cache results of the function `func`.

    This uses the function :func:`_cache_key`.

    EXAMPLES::

        sage: from sage.graphs.tutte_polynomial import tutte_polynomial
        sage: G = graphs.PetersenGraph()
        sage: T = tutte_polynomial(G)  #indirect doctest
        sage: tutte_polynomial(G)(1,1)  #indirect doctest
        2000
    """
    @sage_wraps(func)
    def wrapper(G, *args, **kwds):
        cache = kwds.setdefault('cache', {})
        key = _cache_key(G)
        if key in cache:
            return cache[key]
        result = func(G, *args, **kwds)
        cache[key] = result
        return result
    wrapper.original_func = func
    return wrapper


####################
# Tutte Polynomial #
####################

@_cached
def tutte_polynomial(G, edge_selector=None, cache=None):
    r"""
    Return the Tutte polynomial of the graph `G`.

    INPUT:

    - ``edge_selector`` -- method (optional); this argument allows the user
      to specify his own heuristic for selecting edges used in the deletion
      contraction recurrence

    - ``cache`` -- (optional) dictionary to cache the Tutte
      polynomials generated in the recursive process.  One will be
      created automatically if not provided.

    EXAMPLES:

    The Tutte polynomial of any tree of order `n` is `x^{n-1}`::

        sage: all(T.tutte_polynomial() == x**9 for T in graphs.trees(10))               # needs sage.symbolic
        True

    The Tutte polynomial of the Petersen graph is::

        sage: P = graphs.PetersenGraph()
        sage: P.tutte_polynomial()
        x^9 + 6*x^8 + 21*x^7 + 56*x^6 + 12*x^5*y + y^6 + 114*x^5 + 70*x^4*y
        + 30*x^3*y^2 + 15*x^2*y^3 + 10*x*y^4 + 9*y^5 + 170*x^4 + 170*x^3*y
        + 105*x^2*y^2 + 65*x*y^3 + 35*y^4 + 180*x^3 + 240*x^2*y + 171*x*y^2
        + 75*y^3 + 120*x^2 + 168*x*y + 84*y^2 + 36*x + 36*y

    The Tutte polynomial of a connected graph `G` evaluated at (1,1) is the number of
    spanning trees of `G`::

        sage: G = graphs.RandomGNP(10,0.6)
        sage: while not G.is_connected():
        ....:     G = graphs.RandomGNP(10,0.6)
        sage: G.tutte_polynomial()(1,1) == G.spanning_trees_count()                     # needs sage.modules
        True

    Given that `T(x,y)` is the Tutte polynomial of a graph `G` with
    `n` vertices and `c` connected components, then `(-1)^{n-c} x^k
    T(1-x,0)` is the chromatic polynomial of `G`. ::

        sage: G = graphs.OctahedralGraph()
        sage: T = G.tutte_polynomial()
        sage: R = PolynomialRing(ZZ, 'x')
        sage: R((-1)^5*x*T(1-x,0)).factor()                                             # needs sage.symbolic
        (x - 2) * (x - 1) * x * (x^3 - 9*x^2 + 29*x - 32)
        sage: G.chromatic_polynomial().factor()                                         # needs sage.libs.flint
        (x - 2) * (x - 1) * x * (x^3 - 9*x^2 + 29*x - 32)

    TESTS:

    Providing an external cache::

        sage: cache = {}
        sage: _ = graphs.RandomGNP(7,.5).tutte_polynomial(cache=cache)
        sage: len(cache) > 0
        True

    Verify that :issue:`18366` is fixed::

        sage: g = Graph(multiedges=True)
        sage: g.add_edges([(0,1,1),(1,5,2),(5,3,3),(5,2,4),(2,4,5),(0,2,6),(0,3,7),(0,4,8),(0,5,9)])
        sage: g.tutte_polynomial()(1,1)
        52
        sage: g.spanning_trees_count()                                                  # needs sage.modules
        52
    """
    R = ZZ['x, y']
    if G.num_edges() == 0:
        return R.one()

    G = G.relabel(inplace=False, immutable=False)  # making sure the vertices are integers
    G.allow_loops(True)
    G.allow_multiple_edges(True)

    if edge_selector is None:
        edge_selector = MinimizeSingleDegree()
    x, y = R.gens()
    return _tutte_polynomial_internal(G, x, y, edge_selector, cache=cache)


@_cached
def _tutte_polynomial_internal(G, x, y, edge_selector, cache=None):
    """
    Do the recursive computation of the Tutte polynomial.

    INPUT:

    - ``G`` -- the graph
    - ``x``, ``y`` -- the variables `x`, `y` respectively
    - ``edge_selector`` -- the heuristic for selecting edges used in the
      deletion contraction recurrence

    TESTS::

        sage: P = graphs.CycleGraph(5)
        sage: P.tutte_polynomial() # indirect doctest
        x^4 + x^3 + x^2 + x + y
    """
    if not G.num_edges():
        return x.parent().one()

    def recursive_tp(graph=None):
        """
        The recursive call -- used so that we do not have to specify
        the same arguments everywhere.
        """
        if graph is None:
            graph = G
        return _tutte_polynomial_internal(graph, x, y, edge_selector, cache=cache)

    # Remove loops
    with removed_loops(G) as loops:
        if loops:
            return y**len(loops) * recursive_tp()

    uG = underlying_graph(G)
    em = edge_multiplicities(G)
    d = list(em.values())

    def yy(start, end):
        return sum(y**i for i in range(start, end+1))

    # Lemma 1
    if G.is_forest():
        return prod(x + yy(1, d_i-1) for d_i in d)

    # Theorem 1: from Haggard, Pearce, Royle 2008
    blocks, cut_vertices = G.blocks_and_cut_vertices()
    if len(blocks) > 1:
        return prod([recursive_tp(G.subgraph(block)) for block in blocks])

    components = G.connected_components_number()
    edge = edge_selector(G)
    unlabeled_edge = edge[:2]

    with removed_edge(G, edge):
        if G.connected_components_number() > components:
            with contracted_edge(G, unlabeled_edge):
                return x*recursive_tp()

    ##################################
    # We are in the biconnected case #
    ##################################

    # Theorem 4: from Haggard, Pearce, and Royle Note that the formula
    # at http://homepages.ecs.vuw.ac.nz/~djp/files/TOMS10.pdf is
    # slightly incorrect.  The initial sum should only go to n-2
    # instead of n (allowing for the last part of the recursion).
    # Additionally, the first operand of the final product should be
    # (x+y^{1...(d_n+d_{n-1}-1)}) instead of just (x+y^(d_n+d_{n-1}-1)
    if uG.num_verts() == uG.num_edges():  # G is a multi-cycle
        n = len(d)
        result = 0
        for i in range(n - 2):
            term = (prod((x + yy(1, d_j-1)) for d_j in d[i+1:]) *
                    prod((yy(0, d_k-1)) for d_k in d[:i]))
            result += term
        # The last part of the recursion
        result += (x + yy(1, d[-1] + d[-2] - 1))*prod(yy(0, d_i-1)
                                                      for d_i in d[:-2])
        return result

    # Theorem 3 from Haggard, Pearce, and Royle, adapted to multi-ears
    ear = Ear.find_ear(uG)
    if ear is not None:
        if (ear.is_cycle and ear.vertices == G.vertices(sort=True)):
            # The graph is an ear (cycle) We should never be in this
            # case since we check for multi-cycles above
            return y + sum(x**i for i in range(1, ear.s))
        else:
            with ear.removed_from(G):
                # result = sum(x^i for i in range(ear.s)) #single ear case
                result = sum((prod(x + yy(1, em[e]-1) for e in ear.unlabeled_edges[i+1:])
                              * prod(yy(0, em[e]-1) for e in ear.unlabeled_edges[:i]))
                             for i in range(len(ear.unlabeled_edges)))
                result *= recursive_tp()

                with contracted_edge(G, [ear.end_points[0],
                                         ear.end_points[-1]]):
                    result += prod(yy(0, em[e]-1)
                                   for e in ear.unlabeled_edges)*recursive_tp()

            return result

    # Theorem 2
    if len(em) == 1:  # the graph is just a multiedge
        return x + sum(y**i for i in range(1, em[unlabeled_edge]))
    else:
        with removed_multiedge(G, unlabeled_edge):
            result = recursive_tp()
            with contracted_edge(G, unlabeled_edge):
                result += sum(y**i for i in range(em[unlabeled_edge]))*recursive_tp()
        return result



# sage_setup: distribution = sagemath-graphs
from sage.graphs.generic_graph_pyx cimport GenericGraph_pyx

cdef class TreelengthConnected:
    cdef unsigned int n
    cdef unsigned short * c_distances
    cdef unsigned short ** distances
    cdef unsigned int diameter
    cdef str name
    cdef dict perm_inv
    cdef bint certificate
    cdef bint k_is_defined
    cdef unsigned int k
    cdef GenericGraph_pyx tree  # The final tree decomposition is stored
    cdef unsigned int length
    cdef bint leq_k
    cdef bint _treelength(self, g, k) noexcept



# sage_setup: distribution = sagemath-graphs
r"""
PQ-Trees

This module implements PQ-Trees, a data structure use to represent all
permutations of the columns of a matrix which satisfy the *consecutive ones*
*property*:

  A binary matrix satisfies the *consecutive ones property* if the 1s are
  contiguous in each of its rows (or equivalently, if no row contains the regexp
  pattern `10^+1`).

  Alternatively, one can say that a sequence of sets `S_1,...,S_n` satisfies the
  *consecutive ones property* if for any `x` the indices of the sets containing
  `x` is an interval of `[1,n]`.

This module is used for the recognition of Interval Graphs (see
:meth:`~sage.graphs.generic_graph.GenericGraph.is_interval`).

**P-tree and Q-tree**


- A `P`-tree with children `c_1,...,c_k` (which can be `P`-trees, `Q`-trees, or
  actual sets of points) indicates that all `k!` permutations of the children
  are allowed.

  Example: `\{1,2\},\{3,4\},\{5,6\}` (disjoint sets can be permuted in any way)

- A `Q`-tree with children `c_1,...,c_k` (which can be `P`-trees, `Q`-trees, or
  actual sets of points) indicates that only two permutations of its children
  are allowed: `c_1,...,c_k` or `c_k,...,c_1`.

  Example: `\{1,2\},\{2,3\},\{3,4\},\{4,5\},\{5,6\}` (only two permutations of
  these sets have the *consecutive ones property*).

**Computation of all possible orderings**

#. In order to compute all permutations of a sequence of sets `S_1,...,S_k`
   satisfying the *consecutive ones property*, we initialize `T` as a `P`-tree
   whose children are all the `S_1,...,S_k`, thus representing the set of all
   `k!` permutations of them.

#. We select some element `x` and update the data structure `T` to restrict the
   permutations it describes to those that keep the occurrences of `x` on an
   interval of `[1,...,k]`. This will result in a new `P`-tree whose children
   are:

   * all `\bar c_x` sets `S_i` which do *not* contain `x`.
   * a new `P`-tree whose children are the `c_x` sets `S_i` containing `x`.

   This describes the set of all `c_x!\times \bar c'_x!` permutations of
   `S_1,...,S_k` that keep the sets containing `x` on an interval.

#. We take a second element `x'` and update the data structure `T` to restrict
   the permutations it describes to those that keep `x'` on an interval of
   `[1,...,k]`. The sets `S_1,...,S_k` belong to 4 categories:

   * The family `S_{00}` of sets which do not contain any of
     `x,x'`.

   * The family `S_{01}` of sets which contain `x'` but do not contain
     `x`.

   * The family `S_{10}` of sets which contain `x` but do not contain
     `x'`.

   * The family `S_{11}` of sets which contain `x'` and `x'`.

   With these notations, the permutations of `S_1,...,S_k` which keep the
   occurrences of `x` and `x'` on an interval are of two forms:

   * <some sets `S_{00}`>, <sets from `S_{10}`>, <sets from `S_{11}`>, <sets from `S_{01}`>, <other sets from `S_{00}`>
   * <some sets `S_{00}`>, <sets from `S_{01}`>, <sets from `S_{11}`>, <sets from `S_{10}`>, <other sets from `S_{00}`>

   These permutations can be modeled with the following `PQ`-tree:

   * A `P`-tree whose children are:

     * All sets from `S_{00}`
     * A `Q`-tree whose children are:

       * A `P`-tree with whose children are the sets from `S_{10}`
       * A `P`-tree with whose children are the sets from `S_{11}`
       * A `P`-tree with whose children are the sets from `S_{01}`

#. One at a time, we update the data structure with each element until they are
   all exhausted, or until we reach a proof that no permutation satisfying the
   *consecutive ones property* exists.

   Using these two types of tree, and exploring the different cases of
   intersection, it is possible to represent all the possible permutations of
   our sets satisfying our constraints, or to prove that no such ordering
   exists. This is the whole purpose of this module, and is explained with more
   details in many places, for example in the following document from Hajiaghayi
   [Haj2000]_.

Authors:

Nathann Cohen (initial implementation)


Methods and functions
---------------------
"""

# ****************************************************************************
#       Copyright (C) 2012 Nathann Cohen <nathann.cohen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# Constants, to make the code more readable

FULL = 2
PARTIAL = 1
EMPTY = 0
ALIGNED = True
UNALIGNED = False


##########################################################################
# Some Helper Functions                                                  #
#                                                                        #
# As the elements of a PQ-Tree can be either P-Trees, Q-Trees, or the    #
# sets themselves (the leaves), the following functions are              #
# meant to be applied both on PQ-Trees and Sets, and mimic for the       #
# latter the behaviour we expect from the corresponding methods          #
# defined in class PQ                                                    #
##########################################################################

def _set_contiguous(tree, x):
    """
    Helper function for updating ``tree``.

    The objective is to ensure that the sets containing ``x`` are contiguous for
    any admissible permutation of its subtrees.

    TESTS::

        sage: from sage.graphs.pq_trees import _set_contiguous, P
        sage: p1 = P([[0, 1], [1, 2], [2, 3], [3, 0]])
        sage: _set_contiguous(p1, 0)
        (1, True)
        sage: p1
        ('P', [{1, 2}, {2, 3}, ('P', [{0, 1}, {0, 3}])])
        sage: _set_contiguous(p1, 2)
        (1, True)
        sage: p1
        ('P', [('P', [{0, 1}, {0, 3}]), ('P', [{1, 2}, {2, 3}])])
        sage: _set_contiguous(p1, 1)
        (1, False)
        sage: p1
        ('P', [('Q', [{0, 3}, {0, 1}, {1, 2}, {2, 3}])])
        sage: p2 = P([[0, 1], [0, 2], [0, 3]])
        sage: _set_contiguous(p2, 0)
        (2, True)
        sage: p2
        ('P', [{0, 1}, {0, 2}, {0, 3}])
        sage: _set_contiguous(p2, Set([1, 2]))
        (0, True)
    """
    if isinstance(tree, PQ):
        return tree.set_contiguous(x)
    elif x in tree:
        return (FULL, ALIGNED)
    return (EMPTY, ALIGNED)


def _new_P(liste):
    """
    Helper function returning a new P-tree.

    TESTS::

        sage: from sage.graphs.pq_trees import _new_P
        sage: _new_P([[1,2], [2,3]])
        ('P', [{1, 2}, {2, 3}])
        sage: _new_P([[1,2]])
        [1, 2]
    """
    if len(liste) > 1:
        return P(liste)
    return liste[0]


def _new_Q(liste):
    """
    Helper function returning a new Q-tree.

    TESTS::

        sage: from sage.graphs.pq_trees import _new_Q
        sage: _new_Q([[1,2], [2,3]])
        ('Q', [{1, 2}, {2, 3}])
        sage: _new_Q([[1,2]])
        [1, 2]
    """
    if len(liste) > 1:
        return Q(liste)
    return liste[0]


def _flatten(x):
    """
    Helper function returning a flatten version of ``x``, if ``x`` is a PQ-tree.

    TESTS::

        sage: from sage.graphs.pq_trees import P, Q, _flatten
        sage: p = Q([P([[1,2], [2,3]])])
        sage: _flatten(p)
        ('P', [{1, 2}, {2, 3}])
        sage: _flatten([p, p])
        [('Q', [('P', [{1, 2}, {2, 3}])]), ('Q', [('P', [{1, 2}, {2, 3}])])]
    """
    if isinstance(x, PQ):
        return x.flatten()
    return x


impossible_msg = "Impossible"


def reorder_sets(sets):
    r"""
    Reorders a collection of sets such that each element appears on an
    interval.

    Given a collection of sets `C = S_1,...,S_k` on a ground set `X`,
    this function attempts to reorder them in such a way that `\forall
    x \in X` and `i<j` with `x\in S_i, S_j`, then `x\in S_l` for every
    `i<l<j` if it exists.

    INPUT:

    - ``sets`` -- list of instances of ``list, Set`` or ``set``

    ALGORITHM: PQ-Trees

    EXAMPLES:

    There is only one way (up to reversal) to represent contiguously
    the sequence of sets `\{i-1, i, i+1\}`::

        sage: from sage.graphs.pq_trees import reorder_sets
        sage: seq = [Set([i-1,i,i+1]) for i in range(1,15)]

    We apply a random permutation::

        sage: p = Permutations(len(seq)).random_element()
        sage: seq = [ seq[p(i+1)-1] for i in range(len(seq)) ]
        sage: ordered = reorder_sets(seq)
        sage: if not 0 in ordered[0]:
        ....:    ordered = ordered.reverse()
        sage: print(ordered)
        [{0, 1, 2}, {1, 2, 3}, {2, 3, 4}, {3, 4, 5}, {4, 5, 6}, {5, 6, 7},
         {8, 6, 7}, {8, 9, 7}, {8, 9, 10}, {9, 10, 11}, {10, 11, 12},
         {11, 12, 13}, {12, 13, 14}, {13, 14, 15}]
    """
    if len(sets) <= 2:
        return sets

    s = set().union(*sets)  # union of the sets

    tree = P(sets)

    for i in s:
        tree.set_contiguous(i)
        tree = _flatten(tree)

    return tree.ordering()


class PQ:
    r"""
    PQ-Trees

    This class should not be instantiated by itself: it is extended by
    :class:`P` and :class:`Q`. See the documentation of
    :mod:`sage.graphs.pq_trees` for more information.

    AUTHOR : Nathann Cohen
    """

    def __init__(self, seq):
        r"""
        Construction of a PQ-Tree.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])

        :issue:`17787`::

            sage: Graph('GvGNp?').is_interval()
            False
        """
        from sage.sets.set import Set

        self._children = []
        for e in seq:
            if isinstance(e, list):
                e = Set(e)

            if e not in self._children:
                self._children.append(e)

    def reverse(self):
        r"""
        Recursively reverse ``self`` and its children.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: p.ordering()
            [{1, 2}, {2, 3}, {2, 4}, {8, 2}, {9, 2}]
            sage: p.reverse()
            sage: p.ordering()
            [{9, 2}, {8, 2}, {2, 4}, {2, 3}, {1, 2}]
        """
        for i in self._children:
            if isinstance(i, PQ):
                i.reverse()

        self._children.reverse()

    def __contains__(self, v):
        r"""
        Test whether there exists an element of ``self`` containing
        an element ``v``.

        INPUT:

        - ``v`` -- an element of the ground set

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: 5 in p
            False
            sage: 9 in p
            True
        """
        return any(v in i for i in self)

    def __iter__(self):
        r"""
        Iterate over the children of ``self``.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: for i in p:
            ....:     print(i)
            {1, 2}
            {2, 3}
            ('P', [{2, 4}, {8, 2}, {9, 2}])
        """
        yield from self._children

    def number_of_children(self):
        r"""
        Return the number of children of ``self``.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: p.number_of_children()
            3
        """
        return len(self._children)

    def ordering(self):
        r"""
        Return the current ordering given by listing the leaves from
        left to right.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: p.ordering()
            [{1, 2}, {2, 3}, {2, 4}, {8, 2}, {9, 2}]
        """
        value = []
        for i in self:
            if isinstance(i, PQ):
                value.extend(i.ordering())
            else:
                value.append(i)

        return value

    def __repr__(self) -> str:
        r"""
        Succinctly represent ``self``.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([[1,2], [2,3], P([[2,4], [2,8], [2,9]])])
            sage: print(p)
            ('Q', [{1, 2}, {2, 3}, ('P', [{2, 4}, {8, 2}, {9, 2}])])
        """
        return str((("P" if isinstance(self, P) else "Q"), self._children))

    def simplify(self, v, left=False, right=False):
        r"""
        Return a simplified copy of ``self`` according to the element ``v``.

        If ``self`` is a partial P-tree for ``v``, we would like to
        restrict the permutations of its children to permutations
        keeping the children containing ``v`` contiguous. This
        function also "locks" all the elements not containing ``v``
        inside a `P`-tree, which is useful when one want to keep the
        elements containing ``v`` on one side (which is the case when
        this method is called).

        INPUT:

        - ``left``, ``right`` -- booleans; whether ``v`` is aligned to the
          right or to the left

        - ``v`` -- an element of the ground set

        OUTPUT:

        If ``self`` is a `Q`-Tree, the sequence of its children is
        returned. If ``self`` is a `P`-tree, 2 `P`-tree are returned,
        namely the two `P`-tree defined above and restricting the
        permutations, in the order implied by ``left, right`` (if
        ``right =True``, the second `P`-tree will be the one gathering
        the elements containing ``v``, if ``left=True``, the
        opposite).

        .. NOTE::

           This method is assumes that ``self`` is partial for ``v``,
           and aligned to the side indicated by ``left, right``.

        EXAMPLES:

        A `P`-Tree ::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = P([[2,4], [1,2], [0,8], [0,5]])
            sage: p.simplify(0, right = True)
            [('P', [{2, 4}, {1, 2}]), ('P', [{0, 8}, {0, 5}])]

        A `Q`-Tree ::

            sage: q = Q([[2,4], [1,2], [0,8], [0,5]])
            sage: q.simplify(0, right = True)
            [{2, 4}, {1, 2}, {0, 8}, {0, 5}]
        """
        if sum([left, right]) != 1:
            raise ValueError("Exactly one of left or right must be specified")

        if isinstance(self, Q):
            L = []
            for c in self._children:
                if (isinstance(c, PQ) and               # Is c partial?
                        v in c and                      # (does c contain sets with
                        any(v not in cc for cc in c)):  # and without v ?)
                    L.extend(c.simplify(v, right=right, left=left))
                else:
                    L.append(c)
            return L
        else:
            empty = []
            full = []
            partial = []

            for c in self._children:
                if v in c:
                    if (isinstance(c, PQ) and               # Is c partial? (does c contain
                            any(v not in cc for cc in c)):  # sets with and without v ?)
                        partial = c.simplify(v, right=right, left=left)
                    else:
                        full.append(c)
                else:
                    empty.append(c)
            if empty:
                empty = [_new_P(empty)]
            if full:
                full = [_new_P(full)]

            if right:
                return empty + partial + full
            else:
                return full + partial + empty

    def flatten(self):
        r"""
        Return a flattened copy of ``self``.

        If ``self`` has only one child, we may as well consider its
        child's children, as ``self`` encodes no information. This
        method recursively "flattens" trees having only on PQ-tree
        child, and returns it.

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = Q([P([[2,4], [2,8], [2,9]])])
            sage: p.flatten()
            ('P', [{2, 4}, {8, 2}, {9, 2}])
        """
        if self.number_of_children() == 1:
            return _flatten(self._children[0])
        else:
            self._children = [_flatten(x) for x in self._children]
            return self


class P(PQ):
    r"""
    A P-Tree is a PQ-Tree whose children can be permuted in any way.

    For more information, see the documentation of :mod:`sage.graphs.pq_trees`.
    """
    def set_contiguous(self, v):
        r"""
        Update ``self`` so that the sets containing ``v`` are
        contiguous for any admissible permutation of its subtrees.

        INPUT:

        - ``v`` -- an element of the ground set

        OUTPUT:

        According to the cases:

            * ``(EMPTY, ALIGNED)`` if no set of the tree contains
              an occurrence of ``v``

            * ``(FULL, ALIGNED)`` if all the sets of the tree contain
              ``v``

            * ``(PARTIAL, ALIGNED)`` if some (but not all) of the sets
              contain ``v``, all of which are aligned
              to the right of the ordering at the end when the function ends

            * ``(PARTIAL, UNALIGNED)`` if some (but not all) of the
              sets contain ``v``, though it is impossible to align them
              all to the right

        In any case, the sets containing ``v`` are contiguous when this
        function ends. If there is no possibility of doing so, the function
        raises a :exc:`ValueError` exception.

        EXAMPLES:

        Ensuring the sets containing ``0`` are continuous::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = P([[0,3], [1,2], [2,3], [2,4], [4,0],[2,8], [2,9]])
            sage: p.set_contiguous(0)
            (1, True)
            sage: print(p)
            ('P', [{1, 2}, {2, 3}, {2, 4}, {8, 2}, {9, 2}, ('P', [{0, 3}, {0, 4}])])

        Impossible situation::

            sage: p = P([[0,1], [1,2], [2,3], [3,0]])
            sage: p.set_contiguous(0)
            (1, True)
            sage: p.set_contiguous(1)
            (1, True)
            sage: p.set_contiguous(2)
            (1, True)
            sage: p.set_contiguous(3)
            Traceback (most recent call last):
            ...
            ValueError: Impossible
        """

        ###############################################################
        # Defining Variables :                                        #
        #                                                             #
        # Collecting the information of which children are FULL of v, #
        # which ones are EMPTY, PARTIAL_ALIGNED and PARTIAL_UNALIGNED #
        #                                                             #
        # Defining variables for their cardinals, just to make the    #
        # code slightly more readable :-)                             #
        ###############################################################

        for x in self:
            _set_contiguous(x, v)
        self.flatten()
        seq = [_set_contiguous(x, v) for x in self]

        f_seq = dict(zip(self, seq))

        set_FULL = []
        set_EMPTY = []
        set_PARTIAL_ALIGNED = []
        set_PARTIAL_UNALIGNED = []

        sorting = {
            (FULL, ALIGNED): set_FULL,
            (EMPTY, ALIGNED): set_EMPTY,
            (PARTIAL, ALIGNED): set_PARTIAL_ALIGNED,
            (PARTIAL, UNALIGNED): set_PARTIAL_UNALIGNED
            }

        for i in self:
            sorting[f_seq[i]].append(i)

        n_FULL = len(set_FULL)
        n_EMPTY = len(set_EMPTY)
        n_PARTIAL_ALIGNED = len(set_PARTIAL_ALIGNED)
        n_PARTIAL_UNALIGNED = len(set_PARTIAL_UNALIGNED)

        # Excludes the situation where there is no solution.
        # read next comment for more explanations

        if (n_PARTIAL_ALIGNED > 2 or
                (n_PARTIAL_UNALIGNED >= 1 and n_EMPTY != self.number_of_children() - 1)):
            raise ValueError(impossible_msg)

        # From now on, there are at most two pq-trees which are partially filled
        # If there is one which is not aligned to the right, all the others are empty

        #########################################################
        # 1/2                                                   #
        #                                                       #
        # Several easy cases where we can decide without paying #
        # attention                                             #
        #########################################################

        # All the children are FULL
        elif n_FULL == self.number_of_children():
            return FULL, True

        # All the children are empty
        elif n_EMPTY == self.number_of_children():
            return EMPTY, True

        # There is a PARTIAL UNALIGNED element (and all the others are
        # empty as we checked before

        elif n_PARTIAL_UNALIGNED == 1:
            return (PARTIAL, UNALIGNED)

        # If there is just one partial element and all the others are
        # empty, we just reorder the set to put it at the right end

        elif (n_PARTIAL_ALIGNED == 1 and
              n_EMPTY == self.number_of_children()-1):

            self._children = set_EMPTY + set_PARTIAL_ALIGNED
            return (PARTIAL, ALIGNED)

        ################################################################
        # 2/2                                                          #
        #                                                              #
        # From now on, there are at most two partial pq-trees and all  #
        # of them have v aligned to their right                        #
        #                                                              #
        # We now want to order them in such a way that all the         #
        # elements containing v are located on the right               #
        ################################################################

        else:

            self._children = []

            # We first move the empty elements to the left, if any

            if n_EMPTY > 0:
                self._children.extend(set_EMPTY)

            # If there is one partial element we but have to add it to
            # the sequence, then add all the full elements

            # We must also make sure these elements will not be
            # reordered in such a way that the elements containing v
            # are not contiguous

            # ==> We create a Q-tree

            if n_PARTIAL_ALIGNED < 2:

                new = []

                # add the partial element, if any
                if n_PARTIAL_ALIGNED == 1:

                    subtree = set_PARTIAL_ALIGNED[0]
                    new.extend(subtree.simplify(v, right=ALIGNED))

                # Then the full elements, if any, in a P-tree (we can
                # permute any two of them while keeping all the
                # elements containing v on an interval

                if n_FULL > 0:

                    new.append(_new_P(set_FULL))

                # We lock all of them in a Q-tree

                self._children.append(_new_Q(new))

                return PARTIAL, True

            # If there are 2 partial elements, we take care of both
            # ends. We also know it will not be possible to align the
            # interval of sets containing v to the right

            else:
                new = []

                # The second partial element is aligned to the right
                # while, as we want to put it at the end of the
                # interval, it should be aligned to the left
                set_PARTIAL_ALIGNED[1].reverse()

                # 1/3
                # Left partial subtree
                subtree = set_PARTIAL_ALIGNED[0]
                new.extend(subtree.simplify(v, right=ALIGNED))

                # 2/3
                # Center (Full elements, in a P-tree, as they can be
                # permuted)

                if n_FULL > 0:
                    new.append(_new_P(set_FULL))

                # 3/3
                # Right partial subtree
                subtree = set_PARTIAL_ALIGNED[1]
                new.extend(subtree.simplify(v, left=ALIGNED))

                # We add all of it, locked in a Q-Tree
                self._children.append(_new_Q(new))

                return PARTIAL, False

    def cardinality(self):
        r"""
        Return the number of orderings allowed by the structure.

        .. SEEALSO::

            :meth:`orderings` -- iterate over all admissible orderings

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = P([[0,3], [1,2], [2,3], [2,4], [4,0],[2,8], [2,9]])
            sage: p.cardinality()
            5040
            sage: p.set_contiguous(3)
            (1, True)
            sage: p.cardinality()
            1440
        """
        from math import factorial
        n = factorial(self.number_of_children())
        for c in self._children:
            if isinstance(c, PQ):
                n = n*c.cardinality()
        return n

    def orderings(self):
        r"""
        Iterate over all orderings of the sets allowed by the structure.

        .. SEEALSO::

            :meth:`cardinality` -- return the number of orderings

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: p = P([[2,4], [1,2], [0,8], [0,5]])
            sage: for o in p.orderings():
            ....:    print(o)
            ({2, 4}, {1, 2}, {0, 8}, {0, 5})
            ({2, 4}, {1, 2}, {0, 5}, {0, 8})
            ({2, 4}, {0, 8}, {1, 2}, {0, 5})
            ({2, 4}, {0, 8}, {0, 5}, {1, 2})
            ...
        """
        from itertools import permutations, product
        for p in permutations(self._children):
            yield from product(*[x.orderings() if isinstance(x, PQ) else [x]
                                 for x in p])


class Q(PQ):
    r"""
    A Q-Tree is a PQ-Tree whose children are ordered up to reversal.

    For more information, see the documentation of :mod:`sage.graphs.pq_trees`.
    """

    def set_contiguous(self, v):
        r"""
        Update ``self`` so that the sets containing ``v`` are
        contiguous for any admissible permutation of its subtrees.

        INPUT:

        - ``v`` -- an element of the ground set

        OUTPUT:

        According to the cases:

            * ``(EMPTY, ALIGNED)`` if no set of the tree contains
              an occurrence of ``v``

            * ``(FULL, ALIGNED)`` if all the sets of the tree contain
              ``v``

            * ``(PARTIAL, ALIGNED)`` if some (but not all) of the sets
              contain ``v``, all of which are aligned
              to the right of the ordering at the end when the function ends

            * ``(PARTIAL, UNALIGNED)`` if some (but not all) of the
              sets contain ``v``, though it is impossible to align them
              all to the right

        In any case, the sets containing ``v`` are contiguous when this
        function ends. If there is no possibility of doing so, the function
        raises a :exc:`ValueError` exception.

        EXAMPLES:

        Ensuring the sets containing ``0`` are continuous::

            sage: from sage.graphs.pq_trees import P, Q
            sage: q = Q([[2,3], Q([[3,0],[3,1]]), Q([[4,0],[4,5]])])
            sage: q.set_contiguous(0)
            (1, False)
            sage: print(q)
            ('Q', [{2, 3}, {1, 3}, {0, 3}, {0, 4}, {4, 5}])

        Impossible situation::

            sage: p = Q([[0,1], [1,2], [2,0]])
            sage: p.set_contiguous(0)
            Traceback (most recent call last):
            ...
            ValueError: Impossible
        """
        #################################################################
        # Guidelines :                                                  #
        #                                                               #
        # As the tree is a Q-Tree, we can but reverse the order in      #
        # which the elements appear. It means that we can but check     #
        # the elements containing v are already contiguous (even        #
        # though we have to take special care of partial elements --    #
        # the endpoints of the interval), and answer accordingly        #
        # (partial, full, empty, aligned..). We also want to align the  #
        # elements containing v to the right if possible.               #
        ################################################################

        ###############################################################
        # Defining Variables :                                        #
        #                                                             #
        # Collecting the information of which children are FULL of v, #
        # which ones are EMPTY, PARTIAL_ALIGNED and PARTIAL_UNALIGNED #
        #                                                             #
        # Defining variables for their cardinals, just to make the    #
        # code slightly more readable :-)                             #
        ###############################################################

        for x in self:
            _set_contiguous(x, v)
        self.flatten()
        seq = [_set_contiguous(x, v) for x in self]

        f_seq = dict(zip(self, seq))

        set_FULL = []
        set_EMPTY = []
        set_PARTIAL_ALIGNED = []
        set_PARTIAL_UNALIGNED = []

        sorting = {
            (FULL, ALIGNED): set_FULL,
            (EMPTY, ALIGNED): set_EMPTY,
            (PARTIAL, ALIGNED): set_PARTIAL_ALIGNED,
            (PARTIAL, UNALIGNED): set_PARTIAL_UNALIGNED
            }

        for i in self:
            sorting[f_seq[i]].append(i)

        n_FULL = len(set_FULL)
        n_EMPTY = len(set_EMPTY)
        n_PARTIAL_ALIGNED = len(set_PARTIAL_ALIGNED)
        n_PARTIAL_UNALIGNED = len(set_PARTIAL_UNALIGNED)

        ###################################################################
        #                                                                 #
        # Picking the good ordering for the children :                    #
        #                                                                 #
        #                                                                 #
        # There is a possibility of aligning to the right iif             #
        # the vector can assume the form (as a regular expression) :      #
        #                                                                 #
        # (EMPTY *) PARTIAL (FULL *) Of course, each of these three       #
        # members could be empty                                          #
        #                                                                 #
        # Hence, in the following case we reverse the vector :            #
        #                                                                 #
        # * if the last element is empty (as we checked the whole         #
        #   vector is not empty                                           #
        #                                                                 #
        # * if the last element is partial, aligned, and all the          #
        #   others are full                                               #
        ###################################################################

        if (f_seq[self._children[-1]] == (EMPTY, ALIGNED) or
            (f_seq[self._children[-1]] == (PARTIAL, ALIGNED) and
             n_FULL == self.number_of_children() - 1)):
            # We reverse the order of the elements in the SET only.
            # Which means that they are still aligned to the right !
            self._children.reverse()

        #########################################################
        # 1/2                                                   #
        #                                                       #
        # Several easy cases where we can decide without paying #
        # attention                                             #
        #########################################################

        # Excludes the situation where there is no solution.
        # read next comment for more explanations

        if (n_PARTIAL_ALIGNED > 2 or
            (n_PARTIAL_UNALIGNED >= 1 and
             n_EMPTY != self.number_of_children() - 1)):
            raise ValueError(impossible_msg)

        # From now on, there are at most two pq-trees which are partially filled
        # If there is one which is not aligned to the right, all the others are empty

        # First trivial case, no checking needed
        elif n_FULL == self.number_of_children():
            return FULL, True

        # Second trivial case, no checking needed
        elif n_EMPTY == self.number_of_children():
            return EMPTY, True

        # Third trivial case, no checking needed
        elif n_PARTIAL_UNALIGNED == 1:
            return (PARTIAL, UNALIGNED)

        # If there is just one partial element
        # and all the others are empty, we just reorder
        # the set to put it at the right end

        elif (n_PARTIAL_ALIGNED == 1 and
              n_EMPTY == self.number_of_children() - 1):

            if set_PARTIAL_ALIGNED[0] == self._children[-1]:
                return (PARTIAL, ALIGNED)

            else:
                return (PARTIAL, UNALIGNED)

        ##############################################################
        # 2/2                                                        #
        #                                                            #
        # We iteratively consider all the children, and check        #
        # that the elements containing v are indeed                  #
        # located on an interval.                                    #
        #                                                            #
        # We are also interested in knowing whether this interval is #
        # aligned to the right                                       #
        #                                                            #
        # Because of the previous tests, we can assume there are at  #
        # most two partial pq-trees and all of them are aligned to   #
        # their right                                                #
        ##############################################################

        else:

            new_children = []

            # Two variables to remember where we are
            # according to the interval

            seen_nonempty = False
            seen_right_end = False

            for i in self:

                type, aligned = f_seq[i]

                # We met an empty element
                if type == EMPTY:

                    # 2 possibilities :
                    #
                    #  * we have NOT met a non-empty element before
                    #    and it just means we are looking at the
                    #    leading empty elements
                    #
                    #  * we have met a non-empty element before and it
                    #    means we will never met another non-empty
                    #    element again => we have seen the right end
                    #    of the interval

                    new_children.append(i)

                    if seen_nonempty:
                        seen_right_end = True

                # We met a non-empty element
                else:
                    if seen_right_end:
                        raise ValueError(impossible_msg)

                    if type == PARTIAL:

                        # if we see an ALIGNED partial tree after
                        # having seen a nonempty element then the
                        # partial tree must be aligned to the left and
                        # so we have seen the right end

                        if seen_nonempty and aligned:
                            i.reverse()
                            seen_right_end = True

                            # right partial subtree
                            subtree = i
                            new_children.extend(subtree.simplify(v, left=True))

                        # If we see an UNALIGNED partial element after
                        # having met a nonempty element, there is no
                        # solution to the alignment problem

                        elif seen_nonempty and not aligned:
                            raise ValueError(impossible_msg)

                        # If we see an unaligned element but no non-empty
                        # element since the beginning, we are witnessing both the
                        # left and right end

                        elif not seen_nonempty and not aligned:
                            raise ValueError("Bon, ben ca arrive O_o")
                            seen_right_end = True

                        elif not seen_nonempty and aligned:

                            # left partial subtree
                            subtree = i

                            new_children.extend(subtree.simplify(v, right=True))

                    else:
                        new_children.append(i)

                    seen_nonempty = True

            # Setting the updated sequence of children
            self._children = new_children

            # Whether we achieved an alignment to the right is the
            # complement of whether we have seen the right end

            return (PARTIAL, not seen_right_end)

    def cardinality(self):
        r"""
        Return the number of orderings allowed by the structure.

        .. SEEALSO::

            :meth:`orderings` -- iterate over all admissible orderings

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: q = Q([[0,3], [1,2], [2,3], [2,4], [4,0],[2,8], [2,9]])
            sage: q.cardinality()
            2
        """
        n = 1
        for c in self._children:
            if isinstance(c, PQ):
                n = n*c.cardinality()

        return n if (self.number_of_children() == 1) else 2*n

    def orderings(self):
        r"""
        Iterate over all orderings of the sets allowed by the structure.

        .. SEEALSO::

            :meth:`cardinality` -- return the number of orderings

        EXAMPLES::

            sage: from sage.graphs.pq_trees import P, Q
            sage: q = Q([[2,4], [1,2], [0,8], [0,5]])
            sage: for o in q.orderings():
            ....:    print(o)
            ({2, 4}, {1, 2}, {0, 8}, {0, 5})
            ({0, 5}, {0, 8}, {1, 2}, {2, 4})
        """
        if len(self._children) == 1:
            c = self._children[0]
            yield from (c.orderings() if isinstance(c, PQ) else [c])
        else:
            from itertools import product
            for o in product(*[x.orderings() if isinstance(x, PQ) else [x]
                               for x in self._children]):
                yield o
                yield o[::-1]
